# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2022 PANTHEON.tech s.r.o.
# Copyright(c) 2022 University of New Hampshire

"""Libraries and utilities for running DPDK Test Suite (DTS).

The various modules in the DTS framework offer:

* Connections to nodes, both interactive and non-interactive,
* A straightforward way to add support for different operating systems of remote nodes,
* Test suite setup, execution and teardown, along with test case setup, execution and teardown,
* Pre-test suite setup and post-test suite teardown.
"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2024 Arm Limited

"""Base interactive shell for DPDK applications.

Provides a base class to create interactive shells based on DPDK.
"""


from abc import ABC
from pathlib import PurePath

from framework.params.eal import EalParams
from framework.remote_session.single_active_interactive_shell import (
    SingleActiveInteractiveShell,
)
from framework.settings import SETTINGS
from framework.testbed_model.cpu import LogicalCoreCount, LogicalCoreList
from framework.testbed_model.sut_node import SutNode


def compute_eal_params(
    sut_node: SutNode,
    params: EalParams | None = None,
    lcore_filter_specifier: LogicalCoreCount | LogicalCoreList = LogicalCoreCount(),
    ascending_cores: bool = True,
    append_prefix_timestamp: bool = True,
) -> EalParams:
    """Compute EAL parameters based on the node's specifications.

    Args:
        sut_node: The SUT node to compute the values for.
        params: If :data:`None`, a new object is created and returned. Otherwise `params.lcore_list`
            is modified according to `lcore_filter_specifier`. A DPDK file prefix is also added. If
            `params.ports` is :data:`None`, then `sut_node.ports` is assigned to it.
        lcore_filter_specifier: A number of lcores/cores/sockets to use or a list of lcore ids to
            use. The default will select one lcore for each of two cores on one socket, in ascending
            order of core ids.
        ascending_cores: Sort cores in ascending order (lowest to highest IDs). If :data:`False`,
            sort in descending order.
        append_prefix_timestamp: If :data:`True`, will append a timestamp to DPDK file prefix.
    """
    if params is None:
        params = EalParams()

    if params.lcore_list is None:
        params.lcore_list = LogicalCoreList(
            sut_node.filter_lcores(lcore_filter_specifier, ascending_cores)
        )

    prefix = params.prefix
    if append_prefix_timestamp:
        prefix = f"{prefix}_{sut_node.dpdk_timestamp}"
    prefix = sut_node.main_session.get_dpdk_file_prefix(prefix)
    if prefix:
        sut_node.dpdk_prefix_list.append(prefix)
    params.prefix = prefix

    if params.allowed_ports is None:
        params.allowed_ports = sut_node.ports

    return params


class DPDKShell(SingleActiveInteractiveShell, ABC):
    """The base class for managing DPDK-based interactive shells.

    This class shouldn't be instantiated directly, but instead be extended.
    It automatically injects computed EAL parameters based on the node in the
    supplied app parameters.
    """

    _node: SutNode
    _app_params: EalParams

    def __init__(
        self,
        node: SutNode,
        privileged: bool = True,
        timeout: float = SETTINGS.timeout,
        lcore_filter_specifier: LogicalCoreCount | LogicalCoreList = LogicalCoreCount(),
        ascending_cores: bool = True,
        append_prefix_timestamp: bool = True,
        app_params: EalParams = EalParams(),
        name: str | None = None,
    ) -> None:
        """Extends :meth:`~.interactive_shell.InteractiveShell.__init__`.

        Adds the `lcore_filter_specifier`, `ascending_cores` and `append_prefix_timestamp` arguments
        which are then used to compute the EAL parameters based on the node's configuration.
        """
        app_params = compute_eal_params(
            node,
            app_params,
            lcore_filter_specifier,
            ascending_cores,
            append_prefix_timestamp,
        )

        super().__init__(node, privileged, timeout, app_params, name)

    def _update_real_path(self, path: PurePath) -> None:
        """Extends :meth:`~.interactive_shell.InteractiveShell._update_real_path`.

        Adds the remote DPDK build directory to the path.
        """
        super()._update_real_path(PurePath(self._node.remote_dpdk_build_dir).joinpath(path))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation
# Copyright(c) 2022-2023 PANTHEON.tech s.r.o.
# Copyright(c) 2022-2023 University of New Hampshire
# Copyright(c) 2024 Arm Limited

"""DTS exceptions.

The exceptions all have different severities expressed as an integer.
The highest severity of all raised exceptions is used as the exit code of DTS.
"""

from enum import IntEnum, unique
from typing import ClassVar


@unique
class ErrorSeverity(IntEnum):
    """The severity of errors that occur during DTS execution.

    All exceptions are caught and the most severe error is used as return code.
    """

    #:
    NO_ERR = 0
    #:
    GENERIC_ERR = 1
    #:
    CONFIG_ERR = 2
    #:
    REMOTE_CMD_EXEC_ERR = 3
    #:
    SSH_ERR = 4
    #:
    INTERNAL_ERR = 5
    #:
    DPDK_BUILD_ERR = 10
    #:
    TESTCASE_VERIFY_ERR = 20
    #:
    BLOCKING_TESTSUITE_ERR = 25


class DTSError(Exception):
    """The base exception from which all DTS exceptions are subclassed.

    Do not use this exception, only use subclassed exceptions.
    """

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.GENERIC_ERR


class SSHConnectionError(DTSError):
    """An unsuccessful SSH connection."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.SSH_ERR
    _host: str
    _errors: list[str]

    def __init__(self, host: str, errors: list[str] | None = None):
        """Define the meaning of the first two arguments.

        Args:
            host: The hostname to which we're trying to connect.
            errors: Any errors that occurred during the connection attempt.
        """
        self._host = host
        self._errors = [] if errors is None else errors

    def __str__(self) -> str:
        """Include the errors in the string representation."""
        message = f"Error trying to connect with {self._host}."
        if self._errors:
            message += f" Errors encountered while retrying: {', '.join(self._errors)}"

        return message


class _SSHTimeoutError(DTSError):
    """The execution of a command via SSH timed out.

    This class is private and meant to be raised as its interactive and non-interactive variants.
    """

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.SSH_ERR
    _command: str

    def __init__(self, command: str):
        """Define the meaning of the first argument.

        Args:
            command: The executed command.
        """
        self._command = command

    def __str__(self) -> str:
        """Add some context to the string representation."""
        return f"{self._command} execution timed out."


class SSHTimeoutError(_SSHTimeoutError):
    """The execution of a command on a non-interactive SSH session timed out."""


class InteractiveSSHTimeoutError(_SSHTimeoutError):
    """The execution of a command on an interactive SSH session timed out."""


class _SSHSessionDeadError(DTSError):
    """The SSH session is no longer alive.

    This class is private and meant to be raised as its interactive and non-interactive variants.
    """

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.SSH_ERR
    _host: str

    def __init__(self, host: str):
        """Define the meaning of the first argument.

        Args:
            host: The hostname of the disconnected node.
        """
        self._host = host

    def __str__(self) -> str:
        """Add some context to the string representation."""
        return f"SSH session with {self._host} has died."


class SSHSessionDeadError(_SSHSessionDeadError):
    """Non-interactive SSH session has died."""


class InteractiveSSHSessionDeadError(_SSHSessionDeadError):
    """Interactive SSH session as died."""


class ConfigurationError(DTSError):
    """An invalid configuration."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.CONFIG_ERR


class RemoteCommandExecutionError(DTSError):
    """An unsuccessful execution of a remote command."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.REMOTE_CMD_EXEC_ERR
    #: The executed command.
    command: str
    _command_stderr: str
    _command_return_code: int

    def __init__(self, command: str, command_stderr: str, command_return_code: int):
        """Define the meaning of the first two arguments.

        Args:
            command: The executed command.
            command_stderr: The stderr of the executed command.
            command_return_code: The return code of the executed command.
        """
        self.command = command
        self._command_stderr = command_stderr
        self._command_return_code = command_return_code

    def __str__(self) -> str:
        """Include the command, its return code and stderr in the string representation."""
        return (
            f"Command '{self.command}' returned a non-zero exit code: "
            f"{self._command_return_code}\nStderr: {self._command_stderr}"
        )


class InteractiveCommandExecutionError(DTSError):
    """An unsuccessful execution of a remote command in an interactive environment."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.REMOTE_CMD_EXEC_ERR


class RemoteFileNotFoundError(DTSError):
    """A remote file or directory is requested but doesn’t exist."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.REMOTE_CMD_EXEC_ERR


class DPDKBuildError(DTSError):
    """A DPDK build failure."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.DPDK_BUILD_ERR


class TestCaseVerifyError(DTSError):
    """A test case failure."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.TESTCASE_VERIFY_ERR


class BlockingTestSuiteError(DTSError):
    """A failure in a blocking test suite."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.BLOCKING_TESTSUITE_ERR
    _suite_name: str

    def __init__(self, suite_name: str) -> None:
        """Define the meaning of the first argument.

        Args:
            suite_name: The blocking test suite.
        """
        self._suite_name = suite_name

    def __str__(self) -> str:
        """Add some context to the string representation."""
        return f"Blocking suite {self._suite_name} failed."


class InternalError(DTSError):
    """An internal error or bug has occurred in DTS."""

    #:
    severity: ClassVar[ErrorSeverity] = ErrorSeverity.INTERNAL_ERR



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2024 PANTHEON.tech s.r.o.

"""Testbed topology representation.

A topology of a testbed captures what links are available between the testbed's nodes.
The link information then implies what type of topology is available.
"""

from dataclasses import dataclass
from os import environ
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING or environ.get("DTS_DOC_BUILD"):
    from enum import Enum as NoAliasEnum
else:
    from aenum import NoAliasEnum

from framework.config import PortConfig
from framework.exception import ConfigurationError

from .port import Port


class TopologyType(int, NoAliasEnum):
    """Supported topology types."""

    #: A topology with no Traffic Generator.
    no_link = 0
    #: A topology with one physical link between the SUT node and the TG node.
    one_link = 1
    #: A topology with two physical links between the Sut node and the TG node.
    two_links = 2
    #: The default topology required by test cases if not specified otherwise.
    default = 2

    @classmethod
    def get_from_value(cls, value: int) -> "TopologyType":
        r"""Get the corresponding instance from value.

        :class:`~enum.Enum`\s that don't allow aliases don't know which instance should be returned
        as there could be multiple valid instances. Except for the :attr:`default` value,
        :class:`TopologyType` is a regular :class:`~enum.Enum`.
        When getting an instance from value, we're not interested in the default,
        since we already know the value, allowing us to remove the ambiguity.
        """
        match value:
            case 0:
                return TopologyType.no_link
            case 1:
                return TopologyType.one_link
            case 2:
                return TopologyType.two_links
            case _:
                raise ConfigurationError("More than two links in a topology are not supported.")


class Topology:
    """Testbed topology.

    The topology contains ports processed into ingress and egress ports.
    If there are no ports on a node, dummy ports (ports with no actual values) are stored.
    If there is only one link available, the ports of this link are stored
    as both ingress and egress ports.

    The dummy ports shouldn't be used. It's up to :class:`~framework.runner.DTSRunner`
    to ensure no test case or suite requiring actual links is executed
    when the topology prohibits it and up to the developers to make sure that test cases
    not requiring any links don't use any ports. Otherwise, the underlying methods
    using the ports will fail.

    Attributes:
        type: The type of the topology.
        tg_port_egress: The egress port of the TG node.
        sut_port_ingress: The ingress port of the SUT node.
        sut_port_egress: The egress port of the SUT node.
        tg_port_ingress: The ingress port of the TG node.
    """

    type: TopologyType
    tg_port_egress: Port
    sut_port_ingress: Port
    sut_port_egress: Port
    tg_port_ingress: Port

    def __init__(self, sut_ports: Iterable[Port], tg_ports: Iterable[Port]):
        """Create the topology from `sut_ports` and `tg_ports`.

        Args:
            sut_ports: The SUT node's ports.
            tg_ports: The TG node's ports.
        """
        port_links = []
        for sut_port in sut_ports:
            for tg_port in tg_ports:
                if (sut_port.identifier, sut_port.peer) == (
                    tg_port.peer,
                    tg_port.identifier,
                ):
                    port_links.append(PortLink(sut_port=sut_port, tg_port=tg_port))

        self.type = TopologyType.get_from_value(len(port_links))
        dummy_port = Port(
            "",
            PortConfig(
                pci="0000:00:00.0",
                os_driver_for_dpdk="",
                os_driver="",
                peer_node="",
                peer_pci="0000:00:00.0",
            ),
        )
        self.tg_port_egress = dummy_port
        self.sut_port_ingress = dummy_port
        self.sut_port_egress = dummy_port
        self.tg_port_ingress = dummy_port
        if self.type > TopologyType.no_link:
            self.tg_port_egress = port_links[0].tg_port
            self.sut_port_ingress = port_links[0].sut_port
            self.sut_port_egress = self.sut_port_ingress
            self.tg_port_ingress = self.tg_port_egress
        if self.type > TopologyType.one_link:
            self.sut_port_egress = port_links[1].sut_port
            self.tg_port_ingress = port_links[1].tg_port


@dataclass(slots=True, frozen=True)
class PortLink:
    """The physical, cabled connection between the ports.

    Attributes:
        sut_port: The port on the SUT node connected to `tg_port`.
        tg_port: The port on the TG node connected to `sut_port`.
    """

    sut_port: Port
    tg_port: Port



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2019 Intel Corporation

import sys


def is_function_line(ln):
    return ln.startswith('\t') and ln.endswith(';\n') and ":" not in ln and "# WINDOWS_NO_EXPORT" not in ln

# MinGW keeps the original .map file but replaces per_lcore* to __emutls_v.per_lcore*
def create_mingw_map_file(input_map, output_map):
    with open(input_map) as f_in, open(output_map, 'w') as f_out:
        f_out.writelines([lines.replace('per_lcore', '__emutls_v.per_lcore') for lines in f_in.readlines()])

def main(args):
    if not args[1].endswith('version.map') or \
            not args[2].endswith('exports.def') and \
            not args[2].endswith('mingw.map'):
        return 1

    if args[2].endswith('mingw.map'):
        create_mingw_map_file(args[1], args[2])
        return 0

# generate def file from map file.
# This works taking indented lines only which end with a ";" and which don't
# have a colon in them, i.e. the lines defining functions only.
    else:
        with open(args[1]) as f_in:
            functions = [ln[:-2] + '\n' for ln in sorted(f_in.readlines())
                         if is_function_line(ln)]
            functions = ["EXPORTS\n"] + functions

    with open(args[2], 'w') as f_out:
        f_out.writelines(functions)
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation
# Copyright(c) 2022-2023 PANTHEON.tech s.r.o.
# Copyright(c) 2022-2023 University of New Hampshire
# Copyright(c) 2024 Arm Limited

"""Common functionality for node management.

A node is any host/server DTS connects to.

The base class, :class:`Node`, provides features common to all nodes and is supposed
to be extended by subclasses with features specific to each node type.
The :func:`~Node.skip_setup` decorator can be used without subclassing.
"""

from abc import ABC

from framework.config import (
    OS,
    DPDKBuildConfiguration,
    NodeConfiguration,
    TestRunConfiguration,
)
from framework.exception import ConfigurationError
from framework.logger import DTSLogger, get_dts_logger

from .cpu import (
    LogicalCore,
    LogicalCoreCount,
    LogicalCoreList,
    LogicalCoreListFilter,
    lcore_filter,
)
from .linux_session import LinuxSession
from .os_session import OSSession
from .port import Port


class Node(ABC):
    """The base class for node management.

    It shouldn't be instantiated, but rather subclassed.
    It implements common methods to manage any node:

        * Connection to the node,
        * Hugepages setup.

    Attributes:
        main_session: The primary OS-aware remote session used to communicate with the node.
        config: The node configuration.
        name: The name of the node.
        lcores: The list of logical cores that DTS can use on the node.
            It's derived from logical cores present on the node and the test run configuration.
        ports: The ports of this node specified in the test run configuration.
    """

    main_session: OSSession
    config: NodeConfiguration
    name: str
    lcores: list[LogicalCore]
    ports: list[Port]
    _logger: DTSLogger
    _other_sessions: list[OSSession]
    _test_run_config: TestRunConfiguration

    def __init__(self, node_config: NodeConfiguration):
        """Connect to the node and gather info during initialization.

        Extra gathered information:

        * The list of available logical CPUs. This is then filtered by
          the ``lcores`` configuration in the YAML test run configuration file,
        * Information about ports from the YAML test run configuration file.

        Args:
            node_config: The node's test run configuration.
        """
        self.config = node_config
        self.name = node_config.name
        self._logger = get_dts_logger(self.name)
        self.main_session = create_session(self.config, self.name, self._logger)

        self._logger.info(f"Connected to node: {self.name}")

        self._get_remote_cpus()
        # filter the node lcores according to the test run configuration
        self.lcores = LogicalCoreListFilter(
            self.lcores, LogicalCoreList(self.config.lcores)
        ).filter()

        self._other_sessions = []
        self._init_ports()

    def _init_ports(self) -> None:
        self.ports = [Port(self.name, port_config) for port_config in self.config.ports]
        self.main_session.update_ports(self.ports)

    def set_up_test_run(
        self,
        test_run_config: TestRunConfiguration,
        dpdk_build_config: DPDKBuildConfiguration,
    ) -> None:
        """Test run setup steps.

        Configure hugepages on all DTS node types. Additional steps can be added by
        extending the method in subclasses with the use of super().

        Args:
            test_run_config: A test run configuration according to which
                the setup steps will be taken.
            dpdk_build_config: The build configuration of DPDK.
        """
        self._setup_hugepages()

    def tear_down_test_run(self) -> None:
        """Test run teardown steps.

        There are currently no common execution teardown steps common to all DTS node types.
        Additional steps can be added by extending the method in subclasses with the use of super().
        """

    def create_session(self, name: str) -> OSSession:
        """Create and return a new OS-aware remote session.

        The returned session won't be used by the node creating it. The session must be used by
        the caller. The session will be maintained for the entire lifecycle of the node object,
        at the end of which the session will be cleaned up automatically.

        Note:
            Any number of these supplementary sessions may be created.

        Args:
            name: The name of the session.

        Returns:
            A new OS-aware remote session.
        """
        session_name = f"{self.name} {name}"
        connection = create_session(
            self.config,
            session_name,
            get_dts_logger(session_name),
        )
        self._other_sessions.append(connection)
        return connection

    def filter_lcores(
        self,
        filter_specifier: LogicalCoreCount | LogicalCoreList,
        ascending: bool = True,
    ) -> list[LogicalCore]:
        """Filter the node's logical cores that DTS can use.

        Logical cores that DTS can use are the ones that are present on the node, but filtered
        according to the test run configuration. The `filter_specifier` will filter cores from
        those logical cores.

        Args:
            filter_specifier: Two different filters can be used, one that specifies the number
                of logical cores per core, cores per socket and the number of sockets,
                and another one that specifies a logical core list.
            ascending: If :data:`True`, use cores with the lowest numerical id first and continue
                in ascending order. If :data:`False`, start with the highest id and continue
                in descending order. This ordering affects which sockets to consider first as well.

        Returns:
            The filtered logical cores.
        """
        self._logger.debug(f"Filtering {filter_specifier} from {self.lcores}.")
        return lcore_filter(
            self.lcores,
            filter_specifier,
            ascending,
        ).filter()

    def _get_remote_cpus(self) -> None:
        """Scan CPUs in the remote OS and store a list of LogicalCores."""
        self._logger.info("Getting CPU information.")
        self.lcores = self.main_session.get_remote_cpus(self.config.use_first_core)

    def _setup_hugepages(self) -> None:
        """Setup hugepages on the node.

        Configure the hugepages only if they're specified in the node's test run configuration.
        """
        if self.config.hugepages:
            self.main_session.setup_hugepages(
                self.config.hugepages.number_of,
                self.main_session.hugepage_size,
                self.config.hugepages.force_first_numa,
            )

    def close(self) -> None:
        """Close all connections and free other resources."""
        if self.main_session:
            self.main_session.close()
        for session in self._other_sessions:
            session.close()


def create_session(node_config: NodeConfiguration, name: str, logger: DTSLogger) -> OSSession:
    """Factory for OS-aware sessions.

    Args:
        node_config: The test run configuration of the node to connect to.
        name: The name of the session.
        logger: The logger instance this session will use.
    """
    match node_config.os:
        case OS.linux:
            return LinuxSession(node_config, name, logger)
        case _:
            raise ConfigurationError(f"Unsupported OS {node_config.os}")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2023 PANTHEON.tech s.r.o.

"""Python interactive shell.

Typical usage example in a TestSuite::

    from framework.remote_session import PythonShell
    python_shell = PythonShell(self.tg_node, timeout=5, privileged=True)
    python_shell.send_command("print('Hello World')")
    python_shell.close()
"""

from pathlib import PurePath
from typing import ClassVar

from .interactive_shell import InteractiveShell


class PythonShell(InteractiveShell):
    """Python interactive shell."""

    #: Python's prompt.
    _default_prompt: ClassVar[str] = ">>>"

    #: This forces the prompt to appear after sending a command.
    _command_extra_chars: ClassVar[str] = "\n"

    #: The Python executable.
    path: ClassVar[PurePath] = PurePath("python3")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Microsoft Corporation
"""This script checks if the system supports huge pages"""

import platform
import ctypes

os_name = platform.system()
if os_name == "Linux":
    try:
        with open("/proc/sys/vm/nr_hugepages") as file_o:
            content = file_o.read()
            print(content)
    except:
        print("0")

elif os_name == "FreeBSD":
    # Assume FreeBSD always has hugepages enabled
    print("1")
elif os_name == "Windows":
    if ctypes.windll.kernel32.GetLargePageMinimum() > 0:
        print("1")
    else:
        print("0")
else:
    print("0")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation
# Copyright(c) 2022-2023 PANTHEON.tech s.r.o.
# Copyright(c) 2022-2023 University of New Hampshire

"""DTS logger module.

The module provides several additional features:

    * The storage of DTS execution stages,
    * Logging to console, a human-readable log file and a machine-readable log file,
    * Optional log files for specific stages.
"""

import logging
from enum import auto
from logging import FileHandler, StreamHandler
from pathlib import Path
from typing import ClassVar

from .utils import StrEnum

date_fmt = "%Y/%m/%d %H:%M:%S"
stream_fmt = "%(asctime)s - %(stage)s - %(name)s - %(levelname)s - %(message)s"
dts_root_logger_name = "dts"


class DtsStage(StrEnum):
    """The DTS execution stage."""

    #:
    pre_run = auto()
    #:
    test_run_setup = auto()
    #:
    test_suite_setup = auto()
    #:
    test_suite = auto()
    #:
    test_suite_teardown = auto()
    #:
    test_run_teardown = auto()
    #:
    post_run = auto()


class DTSLogger(logging.Logger):
    """The DTS logger class.

    The class extends the :class:`~logging.Logger` class to add the DTS execution stage information
    to log records. The stage is common to all loggers, so it's stored in a class variable.

    Any time we switch to a new stage, we have the ability to log to an additional log file along
    with a supplementary log file with machine-readable format. These two log files are used until
    a new stage switch occurs. This is useful mainly for logging per test suite.
    """

    _stage: ClassVar[DtsStage] = DtsStage.pre_run
    _extra_file_handlers: list[FileHandler] = []

    def __init__(self, *args, **kwargs):
        """Extend the constructor with extra file handlers."""
        self._extra_file_handlers = []
        super().__init__(*args, **kwargs)

    def makeRecord(self, *args, **kwargs) -> logging.LogRecord:
        """Generates a record with additional stage information.

        This is the default method for the :class:`~logging.Logger` class. We extend it
        to add stage information to the record.

        :meta private:

        Returns:
            record: The generated record with the stage information.
        """
        record = super().makeRecord(*args, **kwargs)
        record.stage = DTSLogger._stage
        return record

    def add_dts_root_logger_handlers(self, verbose: bool, output_dir: str) -> None:
        """Add logger handlers to the DTS root logger.

        This method should be called only on the DTS root logger.
        The log records from child loggers will propagate to these handlers.

        Three handlers are added:

            * A console handler,
            * A file handler,
            * A supplementary file handler with machine-readable logs
              containing more debug information.

        All log messages will be logged to files. The log level of the console handler
        is configurable with `verbose`.

        Args:
            verbose: If :data:`True`, log all messages to the console.
                If :data:`False`, log to console with the :data:`logging.INFO` level.
            output_dir: The directory where the log files will be located.
                The names of the log files correspond to the name of the logger instance.
        """
        self.setLevel(1)

        sh = StreamHandler()
        sh.setFormatter(logging.Formatter(stream_fmt, date_fmt))
        if not verbose:
            sh.setLevel(logging.INFO)
        self.addHandler(sh)

        self._add_file_handlers(Path(output_dir, self.name))

    def set_stage(self, stage: DtsStage, log_file_path: Path | None = None) -> None:
        """Set the DTS execution stage and optionally log to files.

        Set the DTS execution stage of the DTSLog class and optionally add
        file handlers to the instance if the log file name is provided.

        The file handlers log all messages. One is a regular human-readable log file and
        the other one is a machine-readable log file with extra debug information.

        Args:
            stage: The DTS stage to set.
            log_file_path: An optional path of the log file to use. This should be a full path
                (either relative or absolute) without suffix (which will be appended).
        """
        self._remove_extra_file_handlers()

        if DTSLogger._stage != stage:
            self.info(f"Moving from stage '{DTSLogger._stage}' to stage '{stage}'.")
            DTSLogger._stage = stage

        if log_file_path:
            self._extra_file_handlers.extend(self._add_file_handlers(log_file_path))

    def _add_file_handlers(self, log_file_path: Path) -> list[FileHandler]:
        """Add file handlers to the DTS root logger.

        Add two type of file handlers:

            * A regular file handler with suffix ".log",
            * A machine-readable file handler with suffix ".verbose.log".
              This format provides extensive information for debugging and detailed analysis.

        Args:
            log_file_path: The full path to the log file without suffix.

        Returns:
            The newly created file handlers.

        """
        fh = FileHandler(f"{log_file_path}.log")
        fh.setFormatter(logging.Formatter(stream_fmt, date_fmt))
        self.addHandler(fh)

        verbose_fh = FileHandler(f"{log_file_path}.verbose.log")
        verbose_fh.setFormatter(
            logging.Formatter(
                "%(asctime)s|%(stage)s|%(name)s|%(levelname)s|%(pathname)s|%(lineno)d|"
                "%(funcName)s|%(process)d|%(thread)d|%(threadName)s|%(message)s",
                datefmt=date_fmt,
            )
        )
        self.addHandler(verbose_fh)

        return [fh, verbose_fh]

    def _remove_extra_file_handlers(self) -> None:
        """Remove any extra file handlers that have been added to the logger."""
        if self._extra_file_handlers:
            for extra_file_handler in self._extra_file_handlers:
                self.removeHandler(extra_file_handler)

            self._extra_file_handlers = []


def get_dts_logger(name: str | None = None) -> DTSLogger:
    """Return a DTS logger instance identified by `name`.

    Args:
        name: If :data:`None`, return the DTS root logger.
            If specified, return a child of the DTS root logger.

    Returns:
         The DTS root logger or a child logger identified by `name`.
    """
    original_logger_class = logging.getLoggerClass()
    logging.setLoggerClass(DTSLogger)
    if name:
        name = f"{dts_root_logger_name}.{name}"
    else:
        name = dts_root_logger_name
    logger = logging.getLogger(name)
    logging.setLoggerClass(original_logger_class)
    return logger  # type: ignore[return-value]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2024 Arm Limited

"""The DPDK device blocklisting test suite.

This testing suite ensures tests the port blocklisting functionality of testpmd.
"""

from framework.remote_session.testpmd_shell import TestPmdShell
from framework.test_suite import TestSuite, func_test
from framework.testbed_model.capability import TopologyType, requires
from framework.testbed_model.port import Port


@requires(topology_type=TopologyType.two_links)
class TestBlocklist(TestSuite):
    """DPDK device blocklisting test suite."""

    def verify_blocklisted_ports(self, ports_to_block: list[Port]):
        """Runs testpmd with the given ports blocklisted and verifies the ports."""
        with TestPmdShell(self.sut_node, allowed_ports=[], blocked_ports=ports_to_block) as testpmd:
            allowlisted_ports = {port.device_name for port in testpmd.show_port_info_all()}
            blocklisted_ports = {port.pci for port in ports_to_block}

            # sanity check
            allowed_len = len(allowlisted_ports - blocklisted_ports)
            self.verify(allowed_len > 0, "At least one port should have been allowed")

            blocked = not allowlisted_ports & blocklisted_ports
            self.verify(blocked, "At least one port was not blocklisted")

    @func_test
    def no_blocklisted(self):
        """Run testpmd with no blocklisted device.

        Steps:
            Run testpmd without specifying allowed or blocked ports.
        Verify:
            That no ports were blocked.
        """
        self.verify_blocklisted_ports([])

    @func_test
    def one_port_blocklisted(self):
        """Run testpmd with one blocklisted port.

        Steps:
            Run testpmd with one only one blocklisted port and allowing all the other ones.
        Verify:
            That the port was successfully blocklisted.
        """
        self.verify_blocklisted_ports(self.sut_node.ports[:1])

    @func_test
    def all_but_one_port_blocklisted(self):
        """Run testpmd with all but one blocklisted port.

        Steps:
            Run testpmd with only one allowed port, blocking all the other ones.
        Verify:
            That all specified ports were successfully blocklisted.
        """
        self.verify_blocklisted_ports(self.sut_node.ports[:-1])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2023 PANTHEON.tech s.r.o.
# Copyright(c) 2023 University of New Hampshire

"""Remote interactive and non-interactive sessions.

This package provides modules for managing remote connections to a remote host (node).

The non-interactive sessions send commands and return their output and exit code.

The interactive sessions open an interactive shell which is continuously open,
allowing it to send and receive data within that particular shell.
"""

from framework.config import NodeConfiguration
from framework.logger import DTSLogger

from .interactive_remote_session import InteractiveRemoteSession
from .remote_session import RemoteSession
from .ssh_session import SSHSession


def create_remote_session(
    node_config: NodeConfiguration, name: str, logger: DTSLogger
) -> RemoteSession:
    """Factory for non-interactive remote sessions.

    The function returns an SSH session, but will be extended if support
    for other protocols is added.

    Args:
        node_config: The test run configuration of the node to connect to.
        name: The name of the session.
        logger: The logger instance this session will use.

    Returns:
        The SSH remote session.
    """
    return SSHSession(node_config, name, logger)


def create_interactive_session(
    node_config: NodeConfiguration, logger: DTSLogger
) -> InteractiveRemoteSession:
    """Factory for interactive remote sessions.

    The function returns an interactive SSH session, but will be extended if support
    for other protocols is added.

    Args:
        node_config: The test run configuration of the node to connect to.
        logger: The logger instance this session will use.

    Returns:
        The interactive SSH remote session.
    """
    return InteractiveRemoteSession(node_config, logger)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 NVIDIA Corporation & Affiliates

"""
Analyzing the mlx5 PMD datapath tracings
"""
import sys
import argparse
import bt2

PFX_TX = "pmd.net.mlx5.tx."
PFX_TX_LEN = len(PFX_TX)


class MlxQueue:
    """Queue container object"""

    def __init__(self):
        self.done_burst = []  # completed bursts
        self.wait_burst = []  # waiting for completion
        self.pq_id = 0

    def log(self, all):
        """Log all queue bursts"""
        for txb in self.done_burst:
            txb.log()
        if all == True:
            for txb in self.wait_burst:
                txb.log()


class MlxMbuf:
    """Packet mbufs container object"""

    def __init__(self):
        self.wqe = 0     # wqe id
        self.ptr = None  # first packet mbuf pointer
        self.len = 0     # packet data length
        self.nseg = 0    # number of segments

    def log(self):
        """Log mbuf"""
        out_txt = "    %X: %u" % (self.ptr, self.len)
        if self.nseg != 1:
            out_txt += " (%d segs)" % self.nseg
        print(out_txt)


class MlxWqe:
    """WQE container object"""

    def __init__(self):
        self.mbuf = []    # list of mbufs in WQE
        self.wait_ts = 0  # preceding wait/push timestamp
        self.comp_ts = 0  # send/recv completion timestamp
        self.opcode = 0

    def log(self):
        """Log WQE"""
        wqe_id = (self.opcode >> 8) & 0xFFFF
        wqe_op = self.opcode & 0xFF
        out_txt = "  %04X: " % wqe_id
        if wqe_op == 0xF:
            out_txt += "WAIT"
        elif wqe_op == 0x29:
            out_txt += "EMPW"
        elif wqe_op == 0xE:
            out_txt += "TSO "
        elif wqe_op == 0xA:
            out_txt += "SEND"
        else:
            out_txt += "0x%02X" % wqe_op
        if self.comp_ts != 0:
            out_txt += " (%d, %d)" % (self.wait_ts, self.comp_ts - self.wait_ts)
        else:
            out_txt += " (%d)" % self.wait_ts
        print(out_txt)
        for mbuf in self.mbuf:
            mbuf.log()

    def comp(self, wqe_id, wqe_ts):
        """Return 0 if WQE in not completedLog WQE"""
        if self.comp_ts != 0:
            return 1
        cur_id = (self.opcode >> 8) & 0xFFFF
        if cur_id > wqe_id:
            cur_id -= wqe_id
            if cur_id <= 0x8000:
                return 0
        else:
            cur_id = wqe_id - cur_id
            if cur_id >= 0x8000:
                return 0
        self.comp_ts = wqe_ts
        return 1


class MlxBurst:
    """Packet burst container object"""

    def __init__(self):
        self.wqes = []    # issued burst WQEs
        self.done = 0     # number of sent/recv packets
        self.req = 0      # requested number of packets
        self.call_ts = 0  # burst routine invocation
        self.done_ts = 0  # burst routine done
        self.queue = None

    def log(self):
        """Log burst"""
        port = self.queue.pq_id >> 16
        queue = self.queue.pq_id & 0xFFFF
        if self.req == 0:
            print(
                "%u: tx(p=%u, q=%u, %u/%u pkts (incomplete)"
                % (self.call_ts, port, queue, self.done, self.req)
            )
        else:
            print(
                "%u: tx(p=%u, q=%u, %u/%u pkts in %u"
                % (
                    self.call_ts,
                    port,
                    queue,
                    self.done,
                    self.req,
                    self.done_ts - self.call_ts,
                )
            )
        for wqe in self.wqes:
            wqe.log()

    def comp(self, wqe_id, wqe_ts):
        """Return 0 if not all of WQEs in burst completed"""
        wlen = len(self.wqes)
        if wlen == 0:
            return 0
        for wqe in self.wqes:
            if wqe.comp(wqe_id, wqe_ts) == 0:
                return 0
        return 1


class MlxTrace:
    """Trace representing object"""

    def __init__(self):
        self.tx_blst = {}  # current Tx bursts per CPU
        self.tx_qlst = {}  # active Tx queues per port/queue
        self.tx_wlst = {}  # wait timestamp list per CPU

    def run(self, msg_it, verbose):
        """Run over gathered tracing data and build database"""
        for msg in msg_it:
            if not isinstance(msg, bt2._EventMessageConst):
                continue
            event = msg.event
            if event.name.startswith(PFX_TX):
                do_tx(msg, self, verbose)
            # Handling of other log event cathegories can be added here
        if verbose:
            print("*** End of raw data dump ***")

    def log(self, all):
        """Log gathered trace database"""
        for pq_id in self.tx_qlst:
            queue = self.tx_qlst.get(pq_id)
            queue.log(all)


def do_tx_entry(msg, trace, verbose):
    """Handle entry Tx busrt"""
    event = msg.event
    cpu_id = event["cpu_id"]
    burst = trace.tx_blst.get(cpu_id)
    if burst is not None:
        # continue existing burst after WAIT
        return
    if verbose > 0:
        print("%u:%X tx_entry(real_time=%u, port_id=%u, queue_id=%u)" %
              (msg.default_clock_snapshot.ns_from_origin, cpu_id,
               event["real_time"], event["port_id"], event["queue_id"]))
    # allocate the new burst and append to the queue
    burst = MlxBurst()
    burst.call_ts = event["real_time"]
    if burst.call_ts == 0:
        burst.call_ts = msg.default_clock_snapshot.ns_from_origin
    trace.tx_blst[cpu_id] = burst
    pq_id = event["port_id"] << 16 | event["queue_id"]
    queue = trace.tx_qlst.get(pq_id)
    if queue is None:
        # queue does not exist - allocate the new one
        queue = MlxQueue()
        queue.pq_id = pq_id
        trace.tx_qlst[pq_id] = queue
    burst.queue = queue
    queue.wait_burst.append(burst)


def do_tx_exit(msg, trace, verbose):
    """Handle exit Tx busrt"""
    event = msg.event
    cpu_id = event["cpu_id"]
    if verbose > 0:
        print("%u:%X tx_exit(real_time=%u, nb_sent=%u, nb_req=%u)" %
              (msg.default_clock_snapshot.ns_from_origin, cpu_id,
               event["real_time"], event["nb_sent"], event["nb_req"]))
    burst = trace.tx_blst.get(cpu_id)
    if burst is None:
        return
    burst.done_ts = event["real_time"]
    if burst.done_ts == 0:
        burst.done_ts = msg.default_clock_snapshot.ns_from_origin
    burst.req = event["nb_req"]
    burst.done = event["nb_sent"]
    trace.tx_blst.pop(cpu_id)


def do_tx_wqe(msg, trace, verbose):
    """Handle WQE record"""
    event = msg.event
    cpu_id = event["cpu_id"]
    if verbose > 1:
        print("%u:%X tx_wqe(real_time=%u, opcode=%08X)" %
              (msg.default_clock_snapshot.ns_from_origin, cpu_id,
               event["real_time"], event["opcode"]))
    burst = trace.tx_blst.get(cpu_id)
    if burst is None:
        return
    wqe = MlxWqe()
    wqe.wait_ts = trace.tx_wlst.get(cpu_id)
    if wqe.wait_ts is None:
        wqe.wait_ts = event["real_time"]
        if wqe.wait_ts == 0:
            wqe.wait_ts = msg.default_clock_snapshot.ns_from_origin
    wqe.opcode = event["opcode"]
    burst.wqes.append(wqe)


def do_tx_wait(msg, trace, verbose):
    """Handle WAIT record"""
    event = msg.event
    cpu_id = event["cpu_id"]
    if verbose > 1:
        print("%u:%X tx_wait(ts=%u)" %
              (msg.default_clock_snapshot.ns_from_origin, cpu_id, event["ts"]))
    trace.tx_wlst[cpu_id] = event["ts"]


def do_tx_push(msg, trace, verbose):
    """Handle WQE push event"""
    event = msg.event
    cpu_id = event["cpu_id"]
    if verbose > 2:
        print("%u:%X tx_push(mbuf=%X, pkt_len=%u, nb_segs=%u, wqe_id=%04X)" %
              (msg.default_clock_snapshot.ns_from_origin, cpu_id, event["mbuf"],
               event["mbuf_pkt_len"], event["mbuf_nb_segs"], event["wqe_id"]))
    burst = trace.tx_blst.get(cpu_id)
    if burst is None:
        return
    if not burst.wqes:
        return
    wqe = burst.wqes[-1]
    mbuf = MlxMbuf()
    mbuf.wqe = event["wqe_id"]
    mbuf.ptr = event["mbuf"]
    mbuf.len = event["mbuf_pkt_len"]
    mbuf.nseg = event["mbuf_nb_segs"]
    wqe.mbuf.append(mbuf)


def do_tx_complete(msg, trace, verbose):
    """Handle send completion event"""
    event = msg.event
    pq_id = event["port_id"] << 16 | event["queue_id"]
    if verbose > 1:
        cpu_id = event["cpu_id"]
        print("%u:%X tx_complete(port_id=%u, queue_id=%u, ts=%u, wqe_id=%04X)" %
              (msg.default_clock_snapshot.ns_from_origin, cpu_id,
               event["port_id"], event["queue_id"], event["ts"], event["wqe_id"]))
    queue = trace.tx_qlst.get(pq_id)
    if queue is None:
        return
    qlen = len(queue.wait_burst)
    if qlen == 0:
        return
    wqe_id = event["wqe_id"]
    wqe_ts = event["ts"]
    rmv = 0
    while rmv < qlen:
        burst = queue.wait_burst[rmv]
        if burst.comp(wqe_id, wqe_ts) == 0:
            break
        rmv += 1
    # move completed burst(s) to done list
    if rmv != 0:
        idx = 0
        while idx < rmv:
            burst = queue.wait_burst[idx]
            queue.done_burst.append(burst)
            idx += 1
        queue.wait_burst = queue.wait_burst[rmv:]


def do_tx(msg, trace, verbose):
    """Handle Tx related records"""
    name = msg.event.name[PFX_TX_LEN:]
    if name == "entry":
        do_tx_entry(msg, trace, verbose)
    elif name == "exit":
        do_tx_exit(msg, trace, verbose)
    elif name == "wqe":
        do_tx_wqe(msg, trace, verbose)
    elif name == "wait":
        do_tx_wait(msg, trace, verbose)
    elif name == "push":
        do_tx_push(msg, trace, verbose)
    elif name == "complete":
        do_tx_complete(msg, trace, verbose)
    else:
        print("Error: unrecognized Tx event name: %s" % msg.event.name, file=sys.stderr)
        raise ValueError()


def main() -> int:
    """Script entry point"""
    try:
        parser = argparse.ArgumentParser()
        parser.add_argument("path", nargs=1, type=str, help="input trace folder")
        parser.add_argument("-a", "--all", nargs="?", default=False, const=True,
                            help="show all the bursts, including incomplete ones")
        parser.add_argument("-v", "--verbose", type=int, nargs="?", default=0, const=2,
                            help="show all the records below specified level")
        args = parser.parse_args()

        mlx_tr = MlxTrace()
        msg_it = bt2.TraceCollectionMessageIterator(args.path)
        mlx_tr.run(msg_it, args.verbose)
        mlx_tr.log(args.all)
        return 0
    except ValueError:
        return -1


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation
# Copyright(c) 2022 PANTHEON.tech s.r.o.
# Copyright(c) 2022 University of New Hampshire

"""The DTS executable."""

from framework import settings


def main() -> None:
    """Set DTS settings, then run DTS.

    The DTS settings are taken from the command line arguments and the environment variables.
    The settings object is stored in the module-level variable settings.SETTINGS which the entire
    framework uses. After importing the module (or the variable), any changes to the variable are
    not going to be reflected without a re-import. This means that the SETTINGS variable must
    be modified before the settings module is imported anywhere else in the framework.
    """
    settings.SETTINGS = settings.get_settings()

    from framework.runner import DTSRunner

    dts = DTSRunner()
    dts.run()


# Main program begins here
if __name__ == "__main__":
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2023 PANTHEON.tech s.r.o.

"""Virtual devices model.

Alongside support for physical hardware, DPDK can create various virtual devices.
"""


class VirtualDevice:
    """Base class for virtual devices used by DPDK.

    Attributes:
        name: The name of the virtual device.
    """

    name: str

    def __init__(self, name: str):
        """Initialize the virtual device.

        Args:
            name: The name of the virtual device.
        """
        self.name = name

    def __str__(self) -> str:
        """This corresponds to the name used for DPDK devices."""
        return self.name



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2024 Arm Limited

"""Basic L2 forwarding test suite.

This testing suites runs basic L2 forwarding on testpmd across multiple different queue sizes.
The forwarding test is performed with several packets being sent at once.
"""

from framework.params.testpmd import EthPeer, SimpleForwardingModes
from framework.remote_session.testpmd_shell import TestPmdShell
from framework.test_suite import TestSuite, func_test
from framework.testbed_model.capability import requires
from framework.testbed_model.cpu import LogicalCoreCount
from framework.testbed_model.topology import TopologyType
from framework.utils import generate_random_packets


@requires(topology_type=TopologyType.two_links)
class TestL2fwd(TestSuite):
    """L2 forwarding test suite."""

    #: The total number of packets to generate and send for forwarding.
    NUMBER_OF_PACKETS_TO_SEND = 50
    #: The payload size to use for the generated packets in bytes.
    PAYLOAD_SIZE = 100

    def set_up_suite(self) -> None:
        """Set up the test suite.

        Setup:
            Generate the random packets that will be sent.
        """
        self.packets = generate_random_packets(self.NUMBER_OF_PACKETS_TO_SEND, self.PAYLOAD_SIZE)

    @func_test
    def l2fwd_integrity(self) -> None:
        """Test the L2 forwarding integrity.

        Test:
            Configure a testpmd shell with a different numbers of queues (1, 2, 4 and 8) per run.
            Start up L2 forwarding, send random packets from the TG and verify they were all
            received back.
        """
        queues = [1, 2, 4, 8]

        with TestPmdShell(
            self.sut_node,
            lcore_filter_specifier=LogicalCoreCount(cores_per_socket=4),
            forward_mode=SimpleForwardingModes.mac,
            eth_peer=[EthPeer(1, self.tg_node.ports[1].mac_address)],
            disable_device_start=True,
        ) as shell:
            for queues_num in queues:
                self._logger.info(f"Testing L2 forwarding with {queues_num} queue(s)")
                shell.set_ports_queues(queues_num)
                shell.start()

                received_packets = self.send_packets_and_capture(self.packets)
                expected_packets = self.get_expected_packets(self.packets)
                self.match_all_packets(expected_packets, received_packets)

                shell.stop()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2022 University of New Hampshire
# Copyright(c) 2023 PANTHEON.tech s.r.o.

"""NIC port model.

Basic port information, such as location (the port are identified by their PCI address on a node),
drivers and address.
"""


from dataclasses import dataclass

from framework.config import PortConfig


@dataclass(slots=True, frozen=True)
class PortIdentifier:
    """The port identifier.

    Attributes:
        node: The node where the port resides.
        pci: The PCI address of the port on `node`.
    """

    node: str
    pci: str


@dataclass(slots=True)
class Port:
    """Physical port on a node.

    The ports are identified by the node they're on and their PCI addresses. The port on the other
    side of the connection is also captured here.
    Each port is serviced by a driver, which may be different for the operating system (`os_driver`)
    and for DPDK (`os_driver_for_dpdk`). For some devices, they are the same, e.g.: ``mlx5_core``.

    Attributes:
        identifier: The PCI address of the port on a node.
        os_driver: The operating system driver name when the operating system controls the port,
            e.g.: ``i40e``.
        os_driver_for_dpdk: The operating system driver name for use with DPDK, e.g.: ``vfio-pci``.
        peer: The identifier of a port this port is connected with.
            The `peer` is on a different node.
        mac_address: The MAC address of the port.
        logical_name: The logical name of the port. Must be discovered.
    """

    identifier: PortIdentifier
    os_driver: str
    os_driver_for_dpdk: str
    peer: PortIdentifier
    mac_address: str = ""
    logical_name: str = ""

    def __init__(self, node_name: str, config: PortConfig):
        """Initialize the port from `node_name` and `config`.

        Args:
            node_name: The name of the port's node.
            config: The test run configuration of the port.
        """
        self.identifier = PortIdentifier(
            node=node_name,
            pci=config.pci,
        )
        self.os_driver = config.os_driver
        self.os_driver_for_dpdk = config.os_driver_for_dpdk
        self.peer = PortIdentifier(node=config.peer_node, pci=config.peer_pci)

    @property
    def node(self) -> str:
        """The node where the port resides."""
        return self.identifier.node

    @property
    def pci(self) -> str:
        """The PCI address of the port."""
        return self.identifier.pci


@dataclass(slots=True, frozen=True)
class PortLink:
    """The physical, cabled connection between the ports.

    Attributes:
        sut_port: The port on the SUT node connected to `tg_port`.
        tg_port: The port on the TG node connected to `sut_port`.
    """

    sut_port: Port
    tg_port: Port



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2010-2014 Intel Corporation
#

import sys
import os
import subprocess
import argparse
import platform

from glob import glob
from os.path import exists, basename
from os.path import join as path_join

# The PCI base class for all devices
network_class = {'Class': '02', 'Vendor': None, 'Device': None,
                 'SVendor': None, 'SDevice': None}
acceleration_class = {'Class': '12', 'Vendor': None, 'Device': None,
                      'SVendor': None, 'SDevice': None}
ifpga_class = {'Class': '12', 'Vendor': '8086', 'Device': '0b30',
               'SVendor': None, 'SDevice': None}
encryption_class = {'Class': '10', 'Vendor': None, 'Device': None,
                    'SVendor': None, 'SDevice': None}
intel_processor_class = {'Class': '0b', 'Vendor': '8086', 'Device': None,
                         'SVendor': None, 'SDevice': None}
cavium_sso = {'Class': '08', 'Vendor': '177d', 'Device': 'a04b,a04d',
              'SVendor': None, 'SDevice': None}
cavium_fpa = {'Class': '08', 'Vendor': '177d', 'Device': 'a053',
              'SVendor': None, 'SDevice': None}
cavium_pkx = {'Class': '08', 'Vendor': '177d', 'Device': 'a0dd,a049',
              'SVendor': None, 'SDevice': None}
cavium_tim = {'Class': '08', 'Vendor': '177d', 'Device': 'a051',
              'SVendor': None, 'SDevice': None}
cavium_zip = {'Class': '12', 'Vendor': '177d', 'Device': 'a037',
              'SVendor': None, 'SDevice': None}
avp_vnic = {'Class': '05', 'Vendor': '1af4', 'Device': '1110',
            'SVendor': None, 'SDevice': None}

cnxk_bphy = {'Class': '08', 'Vendor': '177d', 'Device': 'a089',
             'SVendor': None, 'SDevice': None}
cnxk_bphy_cgx = {'Class': '08', 'Vendor': '177d', 'Device': 'a059,a060',
                 'SVendor': None, 'SDevice': None}
cnxk_dma = {'Class': '08', 'Vendor': '177d', 'Device': 'a081',
            'SVendor': None, 'SDevice': None}
cnxk_inl_dev = {'Class': '08', 'Vendor': '177d', 'Device': 'a0f0,a0f1',
                'SVendor': None, 'SDevice': None}

hisilicon_dma = {'Class': '08', 'Vendor': '19e5', 'Device': 'a122',
                 'SVendor': None, 'SDevice': None}
odm_dma = {'Class': '08', 'Vendor': '177d', 'Device': 'a08c',
           'SVendor': None, 'SDevice': None}

intel_dlb = {'Class': '0b', 'Vendor': '8086', 'Device': '270b,2710,2714',
             'SVendor': None, 'SDevice': None}
intel_ioat_bdw = {'Class': '08', 'Vendor': '8086',
                  'Device': '6f20,6f21,6f22,6f23,6f24,6f25,6f26,6f27,6f2e,6f2f',
                  'SVendor': None, 'SDevice': None}
intel_ioat_skx = {'Class': '08', 'Vendor': '8086', 'Device': '2021',
                  'SVendor': None, 'SDevice': None}
intel_ioat_icx = {'Class': '08', 'Vendor': '8086', 'Device': '0b00',
                  'SVendor': None, 'SDevice': None}
intel_idxd_spr = {'Class': '08', 'Vendor': '8086', 'Device': '0b25',
                  'SVendor': None, 'SDevice': None}
intel_idxd_gnrd = {'Class': '08', 'Vendor': '8086', 'Device': '11fb',
                  'SVendor': None, 'SDevice': None}
intel_idxd_dmr = {'Class': '08', 'Vendor': '8086', 'Device': '1212',
                  'SVendor': None, 'SDevice': None}
intel_ntb_skx = {'Class': '06', 'Vendor': '8086', 'Device': '201c',
                 'SVendor': None, 'SDevice': None}
intel_ntb_icx = {'Class': '06', 'Vendor': '8086', 'Device': '347e',
                 'SVendor': None, 'SDevice': None}

cnxk_sso = {'Class': '08', 'Vendor': '177d', 'Device': 'a0f9,a0fa',
            'SVendor': None, 'SDevice': None}
cnxk_npa = {'Class': '08', 'Vendor': '177d', 'Device': 'a0fb,a0fc',
            'SVendor': None, 'SDevice': None}
cn9k_ree = {'Class': '08', 'Vendor': '177d', 'Device': 'a0f4',
            'SVendor': None, 'SDevice': None}

virtio_blk = {'Class': '01', 'Vendor': "1af4", 'Device': '1001,1042',
              'SVendor': None, 'SDevice': None}

cnxk_ml = {'Class': '08', 'Vendor': '177d', 'Device': 'a092',
           'SVendor': None, 'SDevice': None}

network_devices = [network_class, cavium_pkx, avp_vnic, ifpga_class]
baseband_devices = [acceleration_class]
crypto_devices = [encryption_class, intel_processor_class]
dma_devices = [cnxk_dma, hisilicon_dma,
               intel_idxd_gnrd, intel_idxd_dmr, intel_idxd_spr,
               intel_ioat_bdw, intel_ioat_icx, intel_ioat_skx,
               odm_dma]
eventdev_devices = [cavium_sso, cavium_tim, intel_dlb, cnxk_sso]
mempool_devices = [cavium_fpa, cnxk_npa]
compress_devices = [cavium_zip]
regex_devices = [cn9k_ree]
ml_devices = [cnxk_ml]
misc_devices = [cnxk_bphy, cnxk_bphy_cgx, cnxk_inl_dev,
                intel_ntb_skx, intel_ntb_icx,
                virtio_blk]

# global dict ethernet devices present. Dictionary indexed by PCI address.
# Each device within this is itself a dictionary of device properties
devices = {}
# list of supported DPDK drivers
dpdk_drivers = ["igb_uio", "vfio-pci", "uio_pci_generic"]
# list of currently loaded kernel modules
loaded_modules = None

# command-line arg flags
b_flag = None
status_flag = False
force_flag = False
noiommu_flag = False
args = []


# check if a specific kernel module is loaded
def module_is_loaded(module):
    global loaded_modules

    if module == 'vfio_pci':
        module = 'vfio-pci'

    if loaded_modules:
        return module in loaded_modules

    # Get list of sysfs modules (both built-in and dynamically loaded)
    sysfs_path = '/sys/module/'

    # Get the list of directories in sysfs_path
    sysfs_mods = [m for m in os.listdir(sysfs_path)
                  if os.path.isdir(os.path.join(sysfs_path, m))]

    # special case for vfio_pci (module is named vfio-pci,
    # but its .ko is named vfio_pci)
    sysfs_mods = [a if a != 'vfio_pci' else 'vfio-pci' for a in sysfs_mods]

    loaded_modules = sysfs_mods

    # add built-in modules as loaded
    release = platform.uname().release
    filename = os.path.join("/lib/modules/", release, "modules.builtin")
    if os.path.exists(filename):
        try:
            with open(filename) as f:
                loaded_modules += [os.path.splitext(os.path.basename(mod))[0] for mod in f]
        except IOError:
            print("Warning: cannot read list of built-in kernel modules")

    return module in loaded_modules


def check_modules():
    '''Checks that igb_uio is loaded'''
    global dpdk_drivers

    # list of supported modules
    mods = [{"Name": driver, "Found": False} for driver in dpdk_drivers]

    # first check if module is loaded
    for mod in mods:
        if module_is_loaded(mod["Name"]):
            mod["Found"] = True

    # check if we have at least one loaded module
    if True not in [mod["Found"] for mod in mods] and b_flag is not None:
        print("Warning: no supported DPDK kernel modules are loaded", file=sys.stderr)

    # change DPDK driver list to only contain drivers that are loaded
    dpdk_drivers = [mod["Name"] for mod in mods if mod["Found"]]


def has_driver(dev_id):
    '''return true if a device is assigned to a driver. False otherwise'''
    return "Driver_str" in devices[dev_id]


def get_pci_device_details(dev_id, probe_lspci):
    '''This function gets additional details for a PCI device'''
    device = {}

    if probe_lspci:
        extra_info = subprocess.check_output(["lspci", "-vmmks", dev_id]).splitlines()
        # parse lspci details
        for line in extra_info:
            if not line:
                continue
            name, value = line.decode("utf8").split("\t", 1)
            name = name.strip(":") + "_str"
            device[name] = value
    # check for a unix interface name
    device["Interface"] = ""
    for base, dirs, _ in os.walk("/sys/bus/pci/devices/%s/" % dev_id):
        if "net" in dirs:
            device["Interface"] = \
                ",".join(os.listdir(os.path.join(base, "net")))
            break
    # check if a port is used for ssh connection
    device["Ssh_if"] = False
    device["Active"] = ""

    return device


def clear_data():
    '''This function clears any old data'''
    global devices
    devices = {}


def get_device_details(devices_type):
    '''This function populates the "devices" dictionary. The keys used are
    the pci addresses (domain:bus:slot.func). The values are themselves
    dictionaries - one for each NIC.'''
    global devices
    global dpdk_drivers

    # first loop through and read details for all devices
    # request machine readable format, with numeric IDs and String
    dev = {}
    dev_lines = subprocess.check_output(["lspci", "-Dvmmnnk"]).splitlines()
    for dev_line in dev_lines:
        if not dev_line:
            if device_type_match(dev, devices_type):
                # Replace "Driver" with "Driver_str" to have consistency of
                # of dictionary key names
                if "Driver" in dev.keys():
                    dev["Driver_str"] = dev.pop("Driver")
                if "Module" in dev.keys():
                    dev["Module_str"] = dev.pop("Module")
                # use dict to make copy of dev
                devices[dev["Slot"]] = dict(dev)
            # Clear previous device's data
            dev = {}
        else:
            name, value = dev_line.decode("utf8").split("\t", 1)
            value_list = value.rsplit(' ', 1)
            if value_list:
                # String stored in <name>_str
                dev[name.rstrip(":") + '_str'] = value_list[0]
            # Numeric IDs
            dev[name.rstrip(":")] = value_list[len(value_list) - 1] \
                .rstrip("]").lstrip("[")

    if devices_type == network_devices:
        # check what is the interface if any for an ssh connection if
        # any to this host, so we can mark it later.
        ssh_if = []
        route = subprocess.check_output(["ip", "-o", "route"])
        # filter out all lines for 169.254 routes
        route = "\n".join(filter(lambda ln: not ln.startswith("169.254"),
                                 route.decode().splitlines()))
        rt_info = route.split()
        for i in range(len(rt_info) - 1):
            if rt_info[i] == "dev":
                ssh_if.append(rt_info[i + 1])

    # based on the basic info, get extended text details
    for d in devices.keys():
        if not device_type_match(devices[d], devices_type):
            continue

        # get additional info and add it to existing data
        devices[d] = devices[d].copy()
        # No need to probe lspci
        devices[d].update(get_pci_device_details(d, False).items())

        if devices_type == network_devices:
            for _if in ssh_if:
                if _if in devices[d]["Interface"].split(","):
                    devices[d]["Ssh_if"] = True
                    devices[d]["Active"] = "*Active*"
                    break

        # add igb_uio to list of supporting modules if needed
        if "Module_str" in devices[d]:
            for driver in dpdk_drivers:
                if driver not in devices[d]["Module_str"]:
                    devices[d]["Module_str"] = \
                        devices[d]["Module_str"] + ",%s" % driver
        else:
            devices[d]["Module_str"] = ",".join(dpdk_drivers)

        # make sure the driver and module strings do not have any duplicates
        if has_driver(d):
            modules = devices[d]["Module_str"].split(",")
            if devices[d]["Driver_str"] in modules:
                modules.remove(devices[d]["Driver_str"])
                devices[d]["Module_str"] = ",".join(modules)


def device_type_match(dev, devices_type):
    for i in range(len(devices_type)):
        param_count = len(
            [x for x in devices_type[i].values() if x is not None])
        match_count = 0
        if dev["Class"][0:2] == devices_type[i]["Class"]:
            match_count = match_count + 1
            for key in devices_type[i].keys():
                if key != 'Class' and devices_type[i][key]:
                    value_list = devices_type[i][key].split(',')
                    for value in value_list:
                        if value.strip(' ') == dev[key]:
                            match_count = match_count + 1
            # count must be the number of non None parameters to match
            if match_count == param_count:
                return True
    return False


def dev_id_from_dev_name(dev_name):
    '''Take a device "name" - a string passed in by user to identify a NIC
    device, and determine the device id - i.e. the domain:bus:slot.func - for
    it, which can then be used to index into the devices array'''

    # check if it's already a suitable index
    if dev_name in devices:
        return dev_name
    # check if it's an index just missing the domain part
    if "0000:" + dev_name in devices:
        return "0000:" + dev_name

    # check if it's an interface name, e.g. eth1
    for d in devices.keys():
        if dev_name in devices[d]["Interface"].split(","):
            return devices[d]["Slot"]
    # if nothing else matches - error
    raise ValueError("Unknown device: %s. "
                     "Please specify device in \"bus:slot.func\" format" % dev_name)


def unbind_one(dev_id, force):
    '''Unbind the device identified by "dev_id" from its current driver'''
    dev = devices[dev_id]
    if not has_driver(dev_id):
        print("Notice: %s %s %s is not currently managed by any driver" %
              (dev["Slot"], dev["Device_str"], dev["Interface"]), file=sys.stderr)
        return

    # prevent us disconnecting ourselves
    if dev["Ssh_if"] and not force:
        print("Warning: routing table indicates that interface %s is active. "
              "Skipping unbind" % dev_id, file=sys.stderr)
        return

    # write to /sys to unbind
    filename = "/sys/bus/pci/drivers/%s/unbind" % dev["Driver_str"]
    try:
        f = open(filename, "a")
    except OSError as err:
        sys.exit("Error: unbind failed for %s - Cannot open %s: %s" %
                 (dev_id, filename, err))
    f.write(dev_id)
    f.close()


def bind_one(dev_id, driver, force):
    '''Bind the device given by "dev_id" to the driver "driver". If the device
    is already bound to a different driver, it will be unbound first'''
    dev = devices[dev_id]
    saved_driver = None  # used to rollback any unbind in case of failure

    # prevent disconnection of our ssh session
    if dev["Ssh_if"] and not force:
        print("Warning: routing table indicates that interface %s is active. "
              "Not modifying" % dev_id, file=sys.stderr)
        return

    # unbind any existing drivers we don't want
    if has_driver(dev_id):
        if dev["Driver_str"] == driver:
            print("Notice: %s already bound to driver %s, skipping" %
                  (dev_id, driver), file=sys.stderr)
            return
        saved_driver = dev["Driver_str"]
        unbind_one(dev_id, force)
        dev["Driver_str"] = ""  # clear driver string

    # For kernels >= 3.15 driver_override can be used to specify the driver
    # for a device rather than relying on the driver to provide a positive
    # match of the device.  The existing process of looking up
    # the vendor and device ID, adding them to the driver new_id,
    # will erroneously bind other devices too which has the additional burden
    # of unbinding those devices
    if driver in dpdk_drivers:
        filename = "/sys/bus/pci/devices/%s/driver_override" % dev_id
        if exists(filename):
            try:
                f = open(filename, "w")
            except OSError as err:
                print("Error: bind failed for %s - Cannot open %s: %s"
                      % (dev_id, filename, err), file=sys.stderr)
                return
            try:
                f.write("%s" % driver)
                f.close()
            except OSError as err:
                print("Error: bind failed for %s - Cannot write driver %s to "
                      "PCI ID: %s" % (dev_id, driver, err), file=sys.stderr)
                return
        # For kernels < 3.15 use new_id to add PCI id's to the driver
        else:
            filename = "/sys/bus/pci/drivers/%s/new_id" % driver
            try:
                f = open(filename, "w")
            except OSError as err:
                print("Error: bind failed for %s - Cannot open %s: %s"
                      % (dev_id, filename, err), file=sys.stderr)
                return
            try:
                # Convert Device and Vendor Id to int to write to new_id
                f.write("%04x %04x" % (int(dev["Vendor"], 16),
                                       int(dev["Device"], 16)))
                f.close()
            except OSError as err:
                print("Error: bind failed for %s - Cannot write new PCI ID to "
                      "driver %s: %s" % (dev_id, driver, err), file=sys.stderr)
                return

    # do the bind by writing to /sys
    filename = "/sys/bus/pci/drivers/%s/bind" % driver
    try:
        f = open(filename, "a")
    except OSError as err:
        print("Error: bind failed for %s - Cannot open %s: %s"
              % (dev_id, filename, err), file=sys.stderr)
        if saved_driver is not None:  # restore any previous driver
            bind_one(dev_id, saved_driver, force)
        return
    try:
        f.write(dev_id)
        f.close()
    except OSError as err:
        # for some reason, closing dev_id after adding a new PCI ID to new_id
        # results in IOError. however, if the device was successfully bound,
        # we don't care for any errors and can safely ignore IOError
        tmp = get_pci_device_details(dev_id, True)
        if "Driver_str" in tmp and tmp["Driver_str"] == driver:
            return
        print("Error: bind failed for %s - Cannot bind to driver %s: %s"
              % (dev_id, driver, err), file=sys.stderr)
        if saved_driver is not None:  # restore any previous driver
            bind_one(dev_id, saved_driver, force)
        return

    # For kernels > 3.15 driver_override is used to bind a device to a driver.
    # Before unbinding it, overwrite driver_override with empty string so that
    # the device can be bound to any other driver
    filename = "/sys/bus/pci/devices/%s/driver_override" % dev_id
    if exists(filename):
        try:
            f = open(filename, "w")
        except OSError as err:
            sys.exit("Error: unbind failed for %s - Cannot open %s: %s"
                     % (dev_id, filename, err))
        try:
            f.write("\00")
            f.close()
        except OSError as err:
            sys.exit("Error: unbind failed for %s - Cannot write %s: %s"
                     % (dev_id, filename, err))


def unbind_all(dev_list, force=False):
    """Unbind method, takes a list of device locations"""

    if dev_list[0] == "dpdk":
        for d in devices.keys():
            if "Driver_str" in devices[d]:
                if devices[d]["Driver_str"] in dpdk_drivers:
                    unbind_one(devices[d]["Slot"], force)
        return

    try:
        dev_list = map(dev_id_from_dev_name, dev_list)
    except ValueError as ex:
        print(ex)
        sys.exit(1)

    for d in dev_list:
        unbind_one(d, force)


def has_iommu():
    """Check if IOMMU is enabled on system"""
    return len(os.listdir("/sys/class/iommu")) > 0


def check_noiommu_mode():
    """Check and enable the noiommu mode for VFIO drivers"""
    global noiommu_flag
    filename = "/sys/module/vfio/parameters/enable_unsafe_noiommu_mode"

    try:
        with open(filename, "r") as f:
            value = f.read(1)
            if value in ("1", "y" ,"Y"):
                return
    except OSError as err:
        sys.exit(f"Error: failed to check unsafe noiommu mode - Cannot open {filename}: {err}")

    if not noiommu_flag:
        sys.exit("Error: IOMMU support is disabled, use --noiommu-mode for binding in noiommu mode")

    try:
        with open(filename, "w") as f:
            f.write("1")
    except OSError as err:
        sys.exit(f"Error: failed to enable unsafe noiommu mode - Cannot open {filename}: {err}")
    print("Warning: enabling unsafe no IOMMU mode for VFIO drivers")


def bind_all(dev_list, driver, force=False):
    """Bind method, takes a list of device locations"""
    global devices

    # a common user error is to forget to specify the driver the devices need to
    # be bound to. check if the driver is a valid device, and if it is, show
    # a meaningful error.
    try:
        dev_id_from_dev_name(driver)
        # if we've made it this far, this means that the "driver" was a valid
        # device string, so it's probably not a valid driver name.
        sys.exit("Error: Driver '%s' does not look like a valid driver. "
                 "Did you forget to specify the driver to bind devices to?" % driver)
    except ValueError:
        # driver generated error - it's not a valid device ID, so all is well
        pass

    # check if we're attempting to bind to a driver that isn't loaded
    if not module_is_loaded(driver.replace('-', '_')):
        sys.exit("Error: Driver '%s' is not loaded." % driver)

    try:
        dev_list = map(dev_id_from_dev_name, dev_list)
    except ValueError as ex:
        sys.exit(ex)

    # check for IOMMU support
    if driver == "vfio-pci" and not has_iommu():
        check_noiommu_mode()

    for d in dev_list:
        bind_one(d, driver, force)

    # For kernels < 3.15 when binding devices to a generic driver
    # (i.e. one that doesn't have a PCI ID table) using new_id, some devices
    # that are not bound to any other driver could be bound even if no one has
    # asked them to. hence, we check the list of drivers again, and see if
    # some of the previously-unbound devices were erroneously bound.
    if not exists("/sys/bus/pci/devices/%s/driver_override" % d):
        for d in devices.keys():
            # skip devices that were already bound or that we know should be bound
            if "Driver_str" in devices[d] or d in dev_list:
                continue

            # update information about this device
            devices[d] = dict(devices[d].items()
                              + get_pci_device_details(d, True).items())

            # check if updated information indicates that the device was bound
            if "Driver_str" in devices[d]:
                unbind_one(d, force)


def display_devices(title, dev_list, extra_params=None):
    '''Displays to the user the details of a list of devices given in
    "dev_list". The "extra_params" parameter, if given, should contain a string
     with %()s fields in it for replacement by the named fields in each
     device's dictionary.'''
    strings = []  # this holds the strings to print. We sort before printing
    print("\n%s" % title)
    print("=" * len(title))
    if not dev_list:
        strings.append("<none>")
    else:
        for dev in dev_list:
            if extra_params is not None:
                strings.append("%s '%s %s' %s" % (dev["Slot"],
                                                  dev["Device_str"],
                                                  dev["Device"],
                                                  extra_params % dev))
            else:
                strings.append("%s '%s'" % (dev["Slot"], dev["Device_str"]))
    # sort before printing, so that the entries appear in PCI order
    strings.sort()
    print("\n".join(strings))  # print one per line


def show_device_status(devices_type, device_name, if_field=False):
    global dpdk_drivers
    kernel_drv = []
    dpdk_drv = []
    no_drv = []

    print_numa = True  # by default, assume we can print NUMA information

    # split our list of network devices into the three categories above
    for d in devices.keys():
        if device_type_match(devices[d], devices_type):
            print_numa &= "NUMANode" in devices[d]
            if not has_driver(d):
                no_drv.append(devices[d])
                continue
            if devices[d]["Driver_str"] in dpdk_drivers:
                dpdk_drv.append(devices[d])
            else:
                kernel_drv.append(devices[d])

    n_devs = len(dpdk_drv) + len(kernel_drv) + len(no_drv)

    # don't bother displaying anything if there are no devices
    if n_devs == 0:
        msg = "No '%s' devices detected" % device_name
        print("")
        print(msg)
        print("".join('=' * len(msg)))
        return

    # print each category separately, so we can clearly see what's used by DPDK
    if dpdk_drv:
        extra_param = "drv=%(Driver_str)s unused=%(Module_str)s"
        if print_numa:
            extra_param = "numa_node=%(NUMANode)s " + extra_param
        display_devices("%s devices using DPDK-compatible driver" % device_name,
                        dpdk_drv, extra_param)
    if kernel_drv:
        extra_param = "drv=%(Driver_str)s unused=%(Module_str)s"
        if if_field:
            extra_param = "if=%(Interface)s " + extra_param
        if print_numa:
            extra_param = "numa_node=%(NUMANode)s " + extra_param
        extra_param += " %(Active)s"
        display_devices("%s devices using kernel driver" % device_name,
                        kernel_drv, extra_param)
    if no_drv:
        extra_param = "unused=%(Module_str)s"
        if print_numa:
            extra_param = "numa_node=%(NUMANode)s " + extra_param
        display_devices("Other %s devices" % device_name, no_drv, extra_param)


def show_status():
    '''Function called when the script is passed the "--status" option.
    Displays to the user what devices are bound to the igb_uio driver, the
    kernel driver or to no driver'''

    if status_dev in ["net", "all"]:
        show_device_status(network_devices, "Network", if_field=True)

    if status_dev in ["baseband", "all"]:
        show_device_status(baseband_devices, "Baseband")

    if status_dev in ["crypto", "all"]:
        show_device_status(crypto_devices, "Crypto")

    if status_dev in ["dma", "all"]:
        show_device_status(dma_devices, "DMA")

    if status_dev in ["event", "all"]:
        show_device_status(eventdev_devices, "Eventdev")

    if status_dev in ["mempool", "all"]:
        show_device_status(mempool_devices, "Mempool")

    if status_dev in ["compress", "all"]:
        show_device_status(compress_devices, "Compress")

    if status_dev in ["misc", "all"]:
        show_device_status(misc_devices, "Misc (rawdev)")

    if status_dev in ["regex", "all"]:
        show_device_status(regex_devices, "Regex")

    if status_dev in ["ml", "all"]:
        show_device_status(ml_devices, "ML")


def pci_glob(arg):
    '''Returns a list containing either:
    * List of PCI B:D:F matching arg, using shell wildcards e.g. 80:04.*
    * Only the passed arg if matching list is empty'''
    sysfs_path = "/sys/bus/pci/devices"
    for _glob in [arg, '0000:' + arg]:
        paths = [basename(path) for path in glob(path_join(sysfs_path, _glob))]
        if paths:
            return paths
    return [arg]


def parse_args():
    '''Parses the command-line arguments given by the user and takes the
    appropriate action for each'''
    global b_flag
    global status_flag
    global status_dev
    global force_flag
    global noiommu_flag
    global args

    parser = argparse.ArgumentParser(
        description='Utility to bind and unbind devices from Linux kernel',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
---------

To display current device status:
        %(prog)s --status

To display current network device status:
        %(prog)s --status-dev net

To bind eth1 from the current driver and move to use vfio-pci
        %(prog)s --bind=vfio-pci eth1

To unbind 0000:01:00.0 from using any driver
        %(prog)s -u 0000:01:00.0

To bind 0000:02:00.0 and 0000:02:00.1 to the ixgbe kernel driver
        %(prog)s -b ixgbe 02:00.0 02:00.1
""")

    parser.add_argument(
        '-s',
        '--status',
        action='store_true',
        help="Print the current status of all known devices.")
    parser.add_argument(
        '--status-dev',
        help="Print the status of given device group.",
        choices=['baseband', 'compress', 'crypto', 'dma', 'event',
                 'mempool', 'misc', 'net', 'regex', 'ml'])
    bind_group = parser.add_mutually_exclusive_group()
    bind_group.add_argument(
        '-b',
        '--bind',
        metavar='DRIVER',
        help="Select the driver to use or \"none\" to unbind the device")
    bind_group.add_argument(
        '-u',
        '--unbind',
        action='store_true',
        help="Unbind a device (equivalent to \"-b none\")")
    parser.add_argument(
        '--noiommu-mode',
        action='store_true',
        help="If IOMMU is not available, enable no IOMMU mode for VFIO drivers")
    parser.add_argument(
        '--force',
        action='store_true',
        help="""
Override restriction on binding devices in use by Linux"
WARNING: This can lead to loss of network connection and should be used with caution.
""")
    parser.add_argument(
        'devices',
        metavar='DEVICE',
        nargs='*',
        help="""
Device specified as PCI "domain:bus:slot.func" syntax or "bus:slot.func" syntax.
For devices bound to Linux kernel drivers, they may be referred to by interface name.
""")

    opt = parser.parse_args()

    if opt.status_dev:
        status_flag = True
        status_dev = opt.status_dev
    if opt.status:
        status_flag = True
        status_dev = "all"
    if opt.force:
        force_flag = True
    if opt.noiommu_mode:
        noiommu_flag = True
    if opt.bind:
        b_flag = opt.bind
    elif opt.unbind:
        b_flag = "none"
    args = opt.devices

    if not b_flag and not status_flag:
        print("Error: No action specified for devices. "
              "Please give a --bind, --ubind or --status option",
              file=sys.stderr)
        parser.print_usage()
        sys.exit(1)

    if b_flag and not args:
        print("Error: No devices specified.", file=sys.stderr)
        parser.print_usage()
        sys.exit(1)

    # resolve any PCI globs in the args
    new_args = []
    for arg in args:
        new_args.extend(pci_glob(arg))
    args = new_args


def do_arg_actions():
    '''do the actual action requested by the user'''
    global b_flag
    global status_flag
    global force_flag
    global args

    if b_flag in ["none", "None"]:
        unbind_all(args, force_flag)
    elif b_flag is not None:
        bind_all(args, b_flag, force_flag)
    if status_flag:
        if b_flag is not None:
            clear_data()
            # refresh if we have changed anything
            get_device_details(network_devices)
            get_device_details(baseband_devices)
            get_device_details(crypto_devices)
            get_device_details(dma_devices)
            get_device_details(eventdev_devices)
            get_device_details(mempool_devices)
            get_device_details(compress_devices)
            get_device_details(regex_devices)
            get_device_details(ml_devices)
            get_device_details(misc_devices)
        show_status()


def main():
    '''program main function'''
    # check if lspci is installed, suppress any output
    with open(os.devnull, 'w') as devnull:
        ret = subprocess.call(['which', 'lspci'],
                              stdout=devnull, stderr=devnull)
        if ret != 0:
            sys.exit("'lspci' not found - please install 'pciutils'")
    parse_args()
    check_modules()
    clear_data()
    get_device_details(network_devices)
    get_device_details(baseband_devices)
    get_device_details(crypto_devices)
    get_device_details(dma_devices)
    get_device_details(eventdev_devices)
    get_device_details(mempool_devices)
    get_device_details(compress_devices)
    get_device_details(regex_devices)
    get_device_details(ml_devices)
    get_device_details(misc_devices)
    do_arg_actions()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2023 Intel Corporation
#
"""
Script to automatically generate boilerplate for using DPDK cmdline library.
"""

import argparse
import shlex
import sys

PARSE_FN_PARAMS = "void *parsed_result, struct cmdline *cl, void *data"
PARSE_FN_BODY = """
    /* TODO: command action */
    RTE_SET_USED(parsed_result);
    RTE_SET_USED(cl);
    RTE_SET_USED(data);
"""
NUMERIC_TYPES = [
    "UINT8",
    "UINT16",
    "UINT32",
    "UINT64",
    "INT8",
    "INT16",
    "INT32",
    "INT64",
]


def process_command(lineno, tokens, comment):
    """Generate the structures and definitions for a single command."""
    out = []
    cfile_out = []

    if tokens[0].startswith("<"):
        raise ValueError(f"Error line {lineno + 1}: command must start with a literal string")

    name_tokens = []
    for t in tokens:
        if t.startswith("<"):
            # stop processing the name building at a variable token,
            # UNLESS the token name starts with "__"
            t_type, t_name = t[1:].split(">")
            if not t_name.startswith("__"):
                break
            t = t_name[2:]   # strip off the leading '__'
        name_tokens.append(t)
    name = "_".join(name_tokens)

    result_struct = []
    initializers = []
    token_list = []
    for t in tokens:
        if t.startswith("<"):
            t_type, t_name = t[1:].split(">")
            t_val = "NULL"
            if t_name.startswith("__"):
                t_name = t_name[2:]
        else:
            t_type = "STRING"
            t_name = t
            t_val = f'"{t}"'

        if t_type == "STRING":
            result_struct.append(f"\tcmdline_fixed_string_t {t_name};")
            initializers.append(
                f"static cmdline_parse_token_string_t cmd_{name}_{t_name}_tok =\n"
                f"\tTOKEN_STRING_INITIALIZER(struct cmd_{name}_result, {t_name}, {t_val});"
            )
        elif t_type in NUMERIC_TYPES:
            result_struct.append(f"\t{t_type.lower()}_t {t_name};")
            initializers.append(
                f"static cmdline_parse_token_num_t cmd_{name}_{t_name}_tok =\n"
                f"\tTOKEN_NUM_INITIALIZER(struct cmd_{name}_result, {t_name}, RTE_{t_type});"
            )
        elif t_type in ["IP", "IP_ADDR", "IPADDR"]:
            result_struct.append(f"\tcmdline_ipaddr_t {t_name};")
            initializers.append(
                f"static cmdline_parse_token_ipaddr_t cmd_{name}_{t_name}_tok =\n"
                f"\tTOKEN_IPADDR_INITIALIZER(struct cmd_{name}_result, {t_name});"
            )
        elif t_type in ["IPV4", "IPv4", "IPV4_ADDR"]:
            result_struct.append(f"\tcmdline_ipaddr_t {t_name};")
            initializers.append(
                f"static cmdline_parse_token_ipaddr_t cmd_{name}_{t_name}_tok =\n"
                f"\tTOKEN_IPV4_INITIALIZER(struct cmd_{name}_result, {t_name});"
            )
        elif t_type in ["IPV6", "IPv6", "IPV6_ADDR"]:
            result_struct.append(f"\tcmdline_ipaddr_t {t_name};")
            initializers.append(
                f"static cmdline_parse_token_ipaddr_t cmd_{name}_{t_name}_tok =\n"
                f"\tTOKEN_IPV6_INITIALIZER(struct cmd_{name}_result, {t_name});"
            )
        elif t_type.startswith("(") and t_type.endswith(")"):
            result_struct.append(f"\tcmdline_fixed_string_t {t_name};")
            t_val = f'"{t_type[1:-1].replace(",","#")}"'
            initializers.append(
                f"static cmdline_parse_token_string_t cmd_{name}_{t_name}_tok =\n"
                f"\tTOKEN_STRING_INITIALIZER(struct cmd_{name}_result, {t_name}, {t_val});"
            )
        else:
            raise TypeError(f"Error line {lineno + 1}: unknown token type '{t_type}'")
        token_list.append(f"cmd_{name}_{t_name}_tok")

    out.append(f'/* Auto-generated handling for command "{" ".join(tokens)}" */')
    # output function prototype
    func_sig = f"void\ncmd_{name}_parsed({PARSE_FN_PARAMS})"
    out.append(f"extern {func_sig};\n")
    # output result data structure
    out.append(f"struct cmd_{name}_result {{\n" + "\n".join(result_struct) + "\n};\n")
    # output the initializer tokens
    out.append("\n".join(initializers) + "\n")
    # output the instance structure
    inst_elems = "\n".join([f"\t\t(void *)&{t}," for t in token_list])
    out.append(
        f"""\
static cmdline_parse_inst_t cmd_{name} = {{
\t.f = cmd_{name}_parsed,
\t.data = NULL,
\t.help_str = "{comment}",
\t.tokens = {{
{inst_elems}
\t\tNULL,
\t}}
}};
"""
    )
    # output function template if C file being written
    cfile_out.append(f"{func_sig}\n{{{PARSE_FN_BODY}}}\n")

    # return the instance structure name
    return (f"cmd_{name}", out, cfile_out)


def process_commands(infile, hfile, cfile, ctxname):
    """Generate boilerplate output for a list of commands from infile."""
    instances = []

    hfile.write(
        f"""\
/* File autogenerated by {sys.argv[0]} */
#ifndef GENERATED_COMMANDS_H
#define GENERATED_COMMANDS_H
#include <rte_common.h>
#include <cmdline.h>
#include <cmdline_parse_string.h>
#include <cmdline_parse_num.h>
#include <cmdline_parse_ipaddr.h>

"""
    )

    for lineno, line in enumerate(infile.readlines()):
        tokens = shlex.split(line, comments=True)
        if not tokens:
            continue
        if "#" in line:
            comment = line.split("#", 1)[-1].strip()
        else:
            comment = ""
        cmd_inst, h_out, c_out = process_command(lineno, tokens, comment)
        hfile.write("\n".join(h_out))
        if cfile:
            cfile.write("\n".join(c_out))
        instances.append(cmd_inst)

    inst_join_str = ",\n\t&"
    hfile.write(
        f"""
static __rte_used cmdline_parse_ctx_t {ctxname}[] = {{
\t&{inst_join_str.join(instances)},
\tNULL
}};

#endif /* GENERATED_COMMANDS_H */
"""
    )


def main():
    """Application main entry point."""
    ap = argparse.ArgumentParser(description=__doc__)
    ap.add_argument(
        "--stubs",
        action="store_true",
        help="Produce C file with empty function stubs for each command",
    )
    ap.add_argument(
        "--output-file",
        "-o",
        default="-",
        help="Output header filename [default to stdout]",
    )
    ap.add_argument(
        "--context-name",
        default="ctx",
        help="Name given to the cmdline context variable in the output header [default=ctx]",
    )
    ap.add_argument("infile", type=argparse.FileType("r"), help="File with list of commands")
    args = ap.parse_args()

    if not args.stubs:
        if args.output_file == "-":
            process_commands(args.infile, sys.stdout, None, args.context_name)
        else:
            with open(args.output_file, "w") as hfile:
                process_commands(args.infile, hfile, None, args.context_name)
    else:
        if not args.output_file.endswith(".h"):
            ap.error(
                "-o/--output-file: specify an output filename ending with .h when creating stubs"
            )

        cfilename = args.output_file[:-2] + ".c"
        with open(args.output_file, "w") as hfile:
            with open(cfilename, "w") as cfile:
                cfile.write(f'#include "{args.output_file}"\n\n')
                process_commands(args.infile, hfile, cfile, args.context_name)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Robin Jarry

MEM_TOTAL = "total_bytes"
MEM_USED = "used_bytes"


def info() -> "dict[Name, tuple[Description, Type]]":
    return {
        MEM_TOTAL: ("The total size of reserved memory in bytes.", "gauge"),
        MEM_USED: ("The currently used memory in bytes.", "gauge"),
    }


def metrics(sock: "TelemetrySocket") -> "list[tuple[Name, Value, Labels]]":
    zones = {}
    used = 0
    for zone in sock.cmd("/eal/memzone_list") or []:
        z = sock.cmd("/eal/memzone_info", zone)
        start = int(z["Hugepage_base"], 16)
        end = start + (z["Hugepage_size"] * z["Hugepage_used"])
        used += z["Length"]
        for s, e in list(zones.items()):
            if s < start < e < end:
                zones[s] = end
                break
            if start < s < end < e:
                del zones[s]
                zones[start] = e
                break
        else:
            zones[start] = end

    return [
        (MEM_TOTAL, sum(end - start for (start, end) in zones.items()), {}),
        (MEM_USED, max(0, used), {}),
    ]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2023 PANTHEON.tech s.r.o.
# Copyright(c) 2023 University of New Hampshire
# Copyright(c) 2024 Arm Limited

"""OS-aware remote session.

DPDK supports multiple different operating systems, meaning it can run on these different operating
systems. This module defines the common API that OS-unaware layers use and translates the API into
OS-aware calls/utility usage.

Note:
    Running commands with administrative privileges requires OS awareness. This is the only layer
    that's aware of OS differences, so this is where non-privileged command get converted
    to privileged commands.

Example:
    A user wishes to remove a directory on a remote :class:`~.sut_node.SutNode`.
    The :class:`~.sut_node.SutNode` object isn't aware what OS the node is running - it delegates
    the OS translation logic to :attr:`~.node.Node.main_session`. The SUT node calls
    :meth:`~OSSession.remove_remote_dir` with a generic, OS-unaware path and
    the :attr:`~.node.Node.main_session` translates that to ``rm -rf`` if the node's OS is Linux
    and other commands for other OSs. It also translates the path to match the underlying OS.
"""
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path, PurePath, PurePosixPath

from framework.config import Architecture, NodeConfiguration
from framework.logger import DTSLogger
from framework.remote_session import (
    InteractiveRemoteSession,
    RemoteSession,
    create_interactive_session,
    create_remote_session,
)
from framework.remote_session.remote_session import CommandResult
from framework.settings import SETTINGS
from framework.utils import MesonArgs, TarCompressionFormat

from .cpu import LogicalCore
from .port import Port


@dataclass(slots=True, frozen=True)
class OSSessionInfo:
    """Supplemental OS session information.

    Attributes:
        os_name: The name of the running operating system of
            the :class:`~framework.testbed_model.node.Node`.
        os_version: The version of the running operating system of
            the :class:`~framework.testbed_model.node.Node`.
        kernel_version: The kernel version of the running operating system of
            the :class:`~framework.testbed_model.node.Node`.
    """

    os_name: str
    os_version: str
    kernel_version: str


class OSSession(ABC):
    """OS-unaware to OS-aware translation API definition.

    The OSSession classes create a remote session to a DTS node and implement OS specific
    behavior. There a few control methods implemented by the base class, the rest need
    to be implemented by subclasses.

    Attributes:
        name: The name of the session.
        remote_session: The remote session maintaining the connection to the node.
        interactive_session: The interactive remote session maintaining the connection to the node.
    """

    _config: NodeConfiguration
    name: str
    _logger: DTSLogger
    remote_session: RemoteSession
    interactive_session: InteractiveRemoteSession
    hugepage_size: int

    def __init__(
        self,
        node_config: NodeConfiguration,
        name: str,
        logger: DTSLogger,
    ):
        """Initialize the OS-aware session.

        Connect to the node right away and also create an interactive remote session.

        Args:
            node_config: The test run configuration of the node to connect to.
            name: The name of the session.
            logger: The logger instance this session will use.
        """
        self.hugepage_size = 2048
        self._config = node_config
        self.name = name
        self._logger = logger
        self.remote_session = create_remote_session(node_config, name, logger)
        self.interactive_session = create_interactive_session(node_config, logger)

    def is_alive(self) -> bool:
        """Check whether the underlying remote session is still responding."""
        return self.remote_session.is_alive()

    def send_command(
        self,
        command: str,
        timeout: float = SETTINGS.timeout,
        privileged: bool = False,
        verify: bool = False,
        env: dict | None = None,
    ) -> CommandResult:
        """An all-purpose API for OS-agnostic commands.

        This can be used for an execution of a portable command that's executed the same way
        on all operating systems, such as Python.

        The :option:`--timeout` command line argument and the :envvar:`DTS_TIMEOUT`
        environment variable configure the timeout of command execution.

        Args:
            command: The command to execute.
            timeout: Wait at most this long in seconds for `command` execution to complete.
            privileged: Whether to run the command with administrative privileges.
            verify: If :data:`True`, will check the exit code of the command.
            env: A dictionary with environment variables to be used with the command execution.

        Raises:
            RemoteCommandExecutionError: If verify is :data:`True` and the command failed.
        """
        if privileged:
            command = self._get_privileged_command(command)

        return self.remote_session.send_command(command, timeout, verify, env)

    def close(self) -> None:
        """Close the underlying remote session."""
        self.remote_session.close()

    @staticmethod
    @abstractmethod
    def _get_privileged_command(command: str) -> str:
        """Modify the command so that it executes with administrative privileges.

        Args:
            command: The command to modify.

        Returns:
            The modified command that executes with administrative privileges.
        """

    @abstractmethod
    def get_remote_tmp_dir(self) -> PurePath:
        """Get the path of the temporary directory of the remote OS.

        Returns:
            The absolute path of the temporary directory.
        """

    @abstractmethod
    def get_dpdk_build_env_vars(self, arch: Architecture) -> dict:
        """Create extra environment variables needed for the target architecture.

        Different architectures may require different configuration, such as setting 32-bit CFLAGS.

        Returns:
            A dictionary with keys as environment variables.
        """

    @abstractmethod
    def join_remote_path(self, *args: str | PurePath) -> PurePath:
        """Join path parts using the path separator that fits the remote OS.

        Args:
            args: Any number of paths to join.

        Returns:
            The resulting joined path.
        """

    @abstractmethod
    def remote_path_exists(self, remote_path: str | PurePath) -> bool:
        """Check whether `remote_path` exists on the remote system.

        Args:
            remote_path: The path to check.

        Returns:
            :data:`True` if the path exists, :data:`False` otherwise.
        """

    @abstractmethod
    def copy_from(self, source_file: str | PurePath, destination_dir: str | Path) -> None:
        """Copy a file from the remote node to the local filesystem.

        Copy `source_file` from the remote node associated with this remote
        session to `destination_dir` on the local filesystem.

        Args:
            source_file: The file on the remote node.
            destination_dir: The directory path on the local filesystem where the `source_file`
                will be saved.
        """

    @abstractmethod
    def copy_to(self, source_file: str | Path, destination_dir: str | PurePath) -> None:
        """Copy a file from local filesystem to the remote node.

        Copy `source_file` from local filesystem to `destination_dir`
        on the remote node associated with this remote session.

        Args:
            source_file: The file on the local filesystem.
            destination_dir: The directory path on the remote Node where the `source_file`
                will be saved.
        """

    @abstractmethod
    def copy_dir_from(
        self,
        source_dir: str | PurePath,
        destination_dir: str | Path,
        compress_format: TarCompressionFormat = TarCompressionFormat.none,
        exclude: str | list[str] | None = None,
    ) -> None:
        """Copy a directory from the remote node to the local filesystem.

        Copy `source_dir` from the remote node associated with this remote session to
        `destination_dir` on the local filesystem. The new local directory will be created
        at `destination_dir` path.

        Example:
            source_dir = '/remote/path/to/source'
            destination_dir = '/local/path/to/destination'
            compress_format = TarCompressionFormat.xz

            The method will:
                1. Create a tarball from `source_dir`, resulting in:
                    '/remote/path/to/source.tar.xz',
                2. Copy '/remote/path/to/source.tar.xz' to
                    '/local/path/to/destination/source.tar.xz',
                3. Extract the contents of the tarball, resulting in:
                    '/local/path/to/destination/source/',
                4. Remove the tarball after extraction
                    ('/local/path/to/destination/source.tar.xz').

            Final Path Structure:
                '/local/path/to/destination/source/'

        Args:
            source_dir: The directory on the remote node.
            destination_dir: The directory path on the local filesystem.
            compress_format: The compression format to use. Defaults to no compression.
            exclude: Patterns for files or directories to exclude from the tarball.
                These patterns are used with `tar`'s `--exclude` option.
        """

    @abstractmethod
    def copy_dir_to(
        self,
        source_dir: str | Path,
        destination_dir: str | PurePath,
        compress_format: TarCompressionFormat = TarCompressionFormat.none,
        exclude: str | list[str] | None = None,
    ) -> None:
        """Copy a directory from the local filesystem to the remote node.

        Copy `source_dir` from the local filesystem to `destination_dir` on the remote node
        associated with this remote session. The new remote directory will be created at
        `destination_dir` path.

        Example:
            source_dir = '/local/path/to/source'
            destination_dir = '/remote/path/to/destination'
            compress_format = TarCompressionFormat.xz

            The method will:
                1. Create a tarball from `source_dir`, resulting in:
                    '/local/path/to/source.tar.xz',
                2. Copy '/local/path/to/source.tar.xz' to
                    '/remote/path/to/destination/source.tar.xz',
                3. Extract the contents of the tarball, resulting in:
                    '/remote/path/to/destination/source/',
                4. Remove the tarball after extraction
                    ('/remote/path/to/destination/source.tar.xz').

            Final Path Structure:
                '/remote/path/to/destination/source/'

        Args:
            source_dir: The directory on the local filesystem.
            destination_dir: The directory path on the remote node.
            compress_format: The compression format to use. Defaults to no compression.
            exclude: Patterns for files or directories to exclude from the tarball.
                These patterns are used with `fnmatch.fnmatch` to filter out files.
        """

    @abstractmethod
    def remove_remote_file(self, remote_file_path: str | PurePath, force: bool = True) -> None:
        """Remove remote file, by default remove forcefully.

        Args:
            remote_file_path: The file path to remove.
            force: If :data:`True`, ignore all warnings and try to remove at all costs.
        """

    @abstractmethod
    def remove_remote_dir(
        self,
        remote_dir_path: str | PurePath,
        recursive: bool = True,
        force: bool = True,
    ) -> None:
        """Remove remote directory, by default remove recursively and forcefully.

        Args:
            remote_dir_path: The directory path to remove.
            recursive: If :data:`True`, also remove all contents inside the directory.
            force: If :data:`True`, ignore all warnings and try to remove at all costs.
        """

    @abstractmethod
    def create_remote_tarball(
        self,
        remote_dir_path: str | PurePath,
        compress_format: TarCompressionFormat = TarCompressionFormat.none,
        exclude: str | list[str] | None = None,
    ) -> PurePosixPath:
        """Create a tarball from the contents of the specified remote directory.

        This method creates a tarball containing all files and directories
        within `remote_dir_path`. The tarball will be saved in the directory of
        `remote_dir_path` and will be named based on `remote_dir_path`.

        Args:
            remote_dir_path: The directory path on the remote node.
            compress_format: The compression format to use. Defaults to no compression.
            exclude: Patterns for files or directories to exclude from the tarball.
                These patterns are used with `tar`'s `--exclude` option.

        Returns:
            The path to the created tarball on the remote node.
        """

    @abstractmethod
    def extract_remote_tarball(
        self,
        remote_tarball_path: str | PurePath,
        expected_dir: str | PurePath | None = None,
    ) -> None:
        """Extract remote tarball in its remote directory.

        Args:
            remote_tarball_path: The tarball path on the remote node.
            expected_dir: If non-empty, check whether `expected_dir` exists after extracting
                the archive.
        """

    @abstractmethod
    def is_remote_dir(self, remote_path: PurePath) -> bool:
        """Check if the `remote_path` is a directory.

        Args:
            remote_tarball_path: The path to the remote tarball.

        Returns:
            If :data:`True` the `remote_path` is a directory, otherwise :data:`False`.
        """

    @abstractmethod
    def is_remote_tarfile(self, remote_tarball_path: PurePath) -> bool:
        """Check if the `remote_tarball_path` is a tar archive.

        Args:
            remote_tarball_path: The path to the remote tarball.

        Returns:
            If :data:`True` the `remote_tarball_path` is a tar archive, otherwise :data:`False`.
        """

    @abstractmethod
    def get_tarball_top_dir(
        self, remote_tarball_path: str | PurePath
    ) -> str | PurePosixPath | None:
        """Get the top directory of the remote tarball.

        Examines the contents of a tarball located at the given `remote_tarball_path` and
        determines the top-level directory. If all files and directories in the tarball share
        the same top-level directory, that directory name is returned. If the tarball contains
        multiple top-level directories or is empty, the method return None.

        Args:
            remote_tarball_path: The path to the remote tarball.

        Returns:
            The top directory of the tarball. If there are multiple top directories
            or the tarball is empty, returns :data:`None`.
        """

    @abstractmethod
    def build_dpdk(
        self,
        env_vars: dict,
        meson_args: MesonArgs,
        remote_dpdk_dir: str | PurePath,
        remote_dpdk_build_dir: str | PurePath,
        rebuild: bool = False,
        timeout: float = SETTINGS.compile_timeout,
    ) -> None:
        """Build DPDK on the remote node.

        An extracted DPDK tarball must be present on the node. The build consists of two steps::

            meson setup <meson args> remote_dpdk_dir remote_dpdk_build_dir
            ninja -C remote_dpdk_build_dir

        The :option:`--compile-timeout` command line argument and the :envvar:`DTS_COMPILE_TIMEOUT`
        environment variable configure the timeout of DPDK build.

        Args:
            env_vars: Use these environment variables when building DPDK.
            meson_args: Use these meson arguments when building DPDK.
            remote_dpdk_dir: The directory on the remote node where DPDK will be built.
            remote_dpdk_build_dir: The target build directory on the remote node.
            rebuild: If :data:`True`, do a subsequent build with ``meson configure`` instead
                of ``meson setup``.
            timeout: Wait at most this long in seconds for the build execution to complete.
        """

    @abstractmethod
    def get_dpdk_version(self, version_path: str | PurePath) -> str:
        """Inspect the DPDK version on the remote node.

        Args:
            version_path: The path to the VERSION file containing the DPDK version.

        Returns:
            The DPDK version.
        """

    @abstractmethod
    def get_remote_cpus(self, use_first_core: bool) -> list[LogicalCore]:
        r"""Get the list of :class:`~.cpu.LogicalCore`\s on the remote node.

        Args:
            use_first_core: If :data:`False`, the first physical core won't be used.

        Returns:
            The logical cores present on the node.
        """

    @abstractmethod
    def kill_cleanup_dpdk_apps(self, dpdk_prefix_list: Iterable[str]) -> None:
        """Kill and cleanup all DPDK apps.

        Args:
            dpdk_prefix_list: Kill all apps identified by `dpdk_prefix_list`.
                If `dpdk_prefix_list` is empty, attempt to find running DPDK apps to kill and clean.
        """

    @abstractmethod
    def get_dpdk_file_prefix(self, dpdk_prefix: str) -> str:
        """Make OS-specific modification to the DPDK file prefix.

        Args:
           dpdk_prefix: The OS-unaware file prefix.

        Returns:
            The OS-specific file prefix.
        """

    @abstractmethod
    def setup_hugepages(self, number_of: int, hugepage_size: int, force_first_numa: bool) -> None:
        """Configure hugepages on the node.

        Get the node's Hugepage Size, configure the specified count of hugepages
        if needed and mount the hugepages if needed.

        Args:
            number_of: Configure this many hugepages.
            hugepage_size: Configure hugepages of this size.
            force_first_numa:  If :data:`True`, configure just on the first numa node.
        """

    @abstractmethod
    def get_compiler_version(self, compiler_name: str) -> str:
        """Get installed version of compiler used for DPDK.

        Args:
            compiler_name: The name of the compiler executable.

        Returns:
            The compiler's version.
        """

    @abstractmethod
    def get_node_info(self) -> OSSessionInfo:
        """Collect additional information about the node.

        Returns:
            Node information.
        """

    @abstractmethod
    def update_ports(self, ports: list[Port]) -> None:
        """Get additional information about ports from the operating system and update them.

        The additional information is:

            * Logical name (e.g. ``enp7s0``) if applicable,
            * Mac address.

        Args:
            ports: The ports to update.
        """

    @abstractmethod
    def configure_port_mtu(self, mtu: int, port: Port) -> None:
        """Configure `mtu` on `port`.

        Args:
            mtu: Desired MTU value.
            port: Port to set `mtu` on.
        """



# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2024 University of New Hampshire

"""Common functionality for interactive shell handling.

The base class, :class:`SingleActiveInteractiveShell`, is meant to be extended by subclasses that
contain functionality specific to that shell type. These subclasses will often modify things like
the prompt to expect or the arguments to pass into the application, but still utilize
the same method for sending a command and collecting output. How this output is handled however
is often application specific. If an application needs elevated privileges to start it is expected
that the method for gaining those privileges is provided when initializing the class.

This class is designed for applications like primary applications in DPDK where only one instance
of the application can be running at a given time and, for this reason, is managed using a context
manager. This context manager starts the application when you enter the context and cleans up the
application when you exit. Using a context manager for this is useful since it allows us to ensure
the application is cleaned up as soon as you leave the block regardless of the reason.

The :option:`--timeout` command line argument and the :envvar:`DTS_TIMEOUT`
environment variable configure the timeout of getting the output from command execution.
"""

from abc import ABC
from pathlib import PurePath
from typing import ClassVar

from paramiko import Channel, channel  # type: ignore[import-untyped]
from typing_extensions import Self

from framework.exception import (
    InteractiveCommandExecutionError,
    InteractiveSSHSessionDeadError,
    InteractiveSSHTimeoutError,
)
from framework.logger import DTSLogger, get_dts_logger
from framework.params import Params
from framework.settings import SETTINGS
from framework.testbed_model.node import Node
from framework.utils import MultiInheritanceBaseClass


class SingleActiveInteractiveShell(MultiInheritanceBaseClass, ABC):
    """The base class for managing interactive shells.

    This class shouldn't be instantiated directly, but instead be extended. It contains
    methods for starting interactive shells as well as sending commands to these shells
    and collecting input until reaching a certain prompt. All interactive applications
    will use the same SSH connection, but each will create their own channel on that
    session.

    Interactive shells are started and stopped using a context manager. This allows for the start
    and cleanup of the application to happen at predictable times regardless of exceptions or
    interrupts.

    Attributes:
        is_alive: :data:`True` if the application has started successfully, :data:`False`
            otherwise.
    """

    _node: Node
    _stdin: channel.ChannelStdinFile
    _stdout: channel.ChannelFile
    _ssh_channel: Channel
    _logger: DTSLogger
    _timeout: float
    _app_params: Params
    _privileged: bool
    _real_path: PurePath

    #: The number of times to try starting the application before considering it a failure.
    _init_attempts: ClassVar[int] = 5

    #: Prompt to expect at the end of output when sending a command.
    #: This is often overridden by subclasses.
    _default_prompt: ClassVar[str] = ""

    #: Extra characters to add to the end of every command
    #: before sending them. This is often overridden by subclasses and is
    #: most commonly an additional newline character. This additional newline
    #: character is used to force the line that is currently awaiting input
    #: into the stdout buffer so that it can be consumed and checked against
    #: the expected prompt.
    _command_extra_chars: ClassVar[str] = ""

    #: Path to the executable to start the interactive application.
    path: ClassVar[PurePath]

    is_alive: bool = False

    def __init__(
        self,
        node: Node,
        privileged: bool = False,
        timeout: float = SETTINGS.timeout,
        app_params: Params = Params(),
        name: str | None = None,
        **kwargs,
    ) -> None:
        """Create an SSH channel during initialization.

        Additional keyword arguments can be passed through `kwargs` if needed for fulfilling other
        constructors in the case of multiple inheritance.

        Args:
            node: The node on which to run start the interactive shell.
            privileged: Enables the shell to run as superuser.
            timeout: The timeout used for the SSH channel that is dedicated to this interactive
                shell. This timeout is for collecting output, so if reading from the buffer
                and no output is gathered within the timeout, an exception is thrown.
            app_params: The command line parameters to be passed to the application on startup.
            name: Name for the interactive shell to use for logging. This name will be appended to
                the name of the underlying node which it is running on.
        """
        self._node = node
        if name is None:
            name = type(self).__name__
        self._logger = get_dts_logger(f"{node.name}.{name}")
        self._app_params = app_params
        self._privileged = privileged
        self._timeout = timeout
        # Ensure path is properly formatted for the host
        self._update_real_path(self.path)
        super().__init__(node, **kwargs)

    def _setup_ssh_channel(self):
        self._ssh_channel = self._node.main_session.interactive_session.session.invoke_shell()
        self._stdin = self._ssh_channel.makefile_stdin("w")
        self._stdout = self._ssh_channel.makefile("r")
        self._ssh_channel.settimeout(self._timeout)
        self._ssh_channel.set_combine_stderr(True)  # combines stdout and stderr streams

    def _make_start_command(self) -> str:
        """Makes the command that starts the interactive shell."""
        start_command = f"{self._real_path} {self._app_params or ''}"
        if self._privileged:
            start_command = self._node.main_session._get_privileged_command(start_command)
        return start_command

    def _start_application(self) -> None:
        """Starts a new interactive application based on the path to the app.

        This method is often overridden by subclasses as their process for starting may look
        different. Initialization of the shell on the host can be retried up to
        `self._init_attempts` - 1 times. This is done because some DPDK applications need slightly
        more time after exiting their script to clean up EAL before others can start.

        Raises:
            InteractiveCommandExecutionError: If the application fails to start within the allotted
                number of retries.
        """
        self._setup_ssh_channel()
        self._ssh_channel.settimeout(5)
        start_command = self._make_start_command()
        self.is_alive = True
        for attempt in range(self._init_attempts):
            try:
                self.send_command(start_command)
                break
            except InteractiveSSHTimeoutError:
                self._logger.info(
                    f"Interactive shell failed to start (attempt {attempt+1} out of "
                    f"{self._init_attempts})"
                )
        else:
            self._ssh_channel.settimeout(self._timeout)
            self.is_alive = False  # update state on failure to start
            raise InteractiveCommandExecutionError("Failed to start application.")
        self._ssh_channel.settimeout(self._timeout)

    def send_command(
        self, command: str, prompt: str | None = None, skip_first_line: bool = False
    ) -> str:
        """Send `command` and get all output before the expected ending string.

        Lines that expect input are not included in the stdout buffer, so they cannot
        be used for expect.

        Example:
            If you were prompted to log into something with a username and password,
            you cannot expect ``username:`` because it won't yet be in the stdout buffer.
            A workaround for this could be consuming an extra newline character to force
            the current `prompt` into the stdout buffer.

        Args:
            command: The command to send.
            prompt: After sending the command, `send_command` will be expecting this string.
                If :data:`None`, will use the class's default prompt.
            skip_first_line: Skip the first line when capturing the output.

        Returns:
            All output in the buffer before expected string.

        Raises:
            InteractiveCommandExecutionError: If attempting to send a command to a shell that is
                not currently running.
            InteractiveSSHSessionDeadError: The session died while executing the command.
            InteractiveSSHTimeoutError: If command was sent but prompt could not be found in
                the output before the timeout.
        """
        if not self.is_alive:
            raise InteractiveCommandExecutionError(
                f"Cannot send command {command} to application because the shell is not running."
            )
        self._logger.info(f"Sending: '{command}'")
        if prompt is None:
            prompt = self._default_prompt
        out: str = ""
        try:
            self._stdin.write(f"{command}{self._command_extra_chars}\n")
            self._stdin.flush()
            for line in self._stdout:
                if skip_first_line:
                    skip_first_line = False
                    continue
                if line.rstrip().endswith(prompt):
                    break
                out += line
        except TimeoutError as e:
            self._logger.exception(e)
            self._logger.debug(
                f"Prompt ({prompt}) was not found in output from command before timeout."
            )
            raise InteractiveSSHTimeoutError(command) from e
        except OSError as e:
            self._logger.exception(e)
            raise InteractiveSSHSessionDeadError(
                self._node.main_session.interactive_session.hostname
            ) from e
        finally:
            self._logger.debug(f"Got output: {out}")
        return out

    def _close(self) -> None:
        self._stdin.close()
        self._ssh_channel.close()

    def _update_real_path(self, path: PurePath) -> None:
        """Updates the interactive shell's real path used at command line."""
        self._real_path = self._node.main_session.join_remote_path(path)

    def __enter__(self) -> Self:
        """Enter the context block.

        Upon entering a context block with this class, the desired behavior is to create the
        channel for the application to use, and then start the application.

        Returns:
            Reference to the object for the application after it has been started.
        """
        self._start_application()
        return self

    def __exit__(self, *_) -> None:
        """Exit the context block.

        Upon exiting a context block with this class, we want to ensure that the instance of the
        application is explicitly closed and properly cleaned up using its close method. Note that
        because this method returns :data:`None` if an exception was raised within the block, it is
        not handled and will be re-raised after the application is closed.

        The desired behavior is to close the application regardless of the reason for exiting the
        context and then recreate that reason afterwards. All method arguments are ignored for
        this reason.
        """
        self._close()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Robin Jarry

RX_PACKETS = "rx_packets"
RX_BYTES = "rx_bytes"
RX_MISSED = "rx_missed"
RX_NOMBUF = "rx_nombuf"
RX_ERRORS = "rx_errors"
TX_PACKETS = "tx_packets"
TX_BYTES = "tx_bytes"
TX_ERRORS = "tx_errors"


def info() -> "dict[Name, tuple[Description, Type]]":
    return {
        RX_PACKETS: ("Number of successfully received packets.", "counter"),
        RX_BYTES: ("Number of successfully received bytes.", "counter"),
        RX_MISSED: (
            "Number of packets dropped by the HW because Rx queues are full.",
            "counter",
        ),
        RX_NOMBUF: ("Number of Rx mbuf allocation failures.", "counter"),
        RX_ERRORS: ("Number of erroneous received packets.", "counter"),
        TX_PACKETS: ("Number of successfully transmitted packets.", "counter"),
        TX_BYTES: ("Number of successfully transmitted bytes.", "counter"),
        TX_ERRORS: ("Number of packet transmission failures.", "counter"),
    }


def metrics(sock: "TelemetrySocket") -> "list[tuple[Name, Value, Labels]]":
    out = []
    for port_id in sock.cmd("/ethdev/list"):
        port = sock.cmd("/ethdev/info", port_id)
        stats = sock.cmd("/ethdev/stats", port_id)
        labels = {"port": port["name"]}
        out += [
            (RX_PACKETS, stats["ipackets"], labels),
            (RX_PACKETS, stats["ipackets"], labels),
            (RX_BYTES, stats["ibytes"], labels),
            (RX_MISSED, stats["imissed"], labels),
            (RX_NOMBUF, stats["rx_nombuf"], labels),
            (RX_ERRORS, stats["ierrors"], labels),
            (TX_PACKETS, stats["opackets"], labels),
            (TX_BYTES, stats["obytes"], labels),
            (TX_ERRORS, stats["oerrors"], labels),
        ]
    return out



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2017 Intel Corporation

import sys
import os
import argparse
import subprocess
import shlex

from threading import Timer

def kill(process):
    print("ERROR: Test app timed out")
    process.kill()

dpdk_path = "../.."
dpdk_target = "build"

parser = argparse.ArgumentParser(
                    description='BBdev Unit Test Application',
                    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("-p", "--testapp-path",
                    help="specifies path to the bbdev test app",
                    default=dpdk_path + "/" + dpdk_target + "/app/dpdk-test-bbdev")
parser.add_argument("-e", "--eal-params",
                    help="EAL arguments which must be passed to the test app",
                    default="--vdev=baseband_null0 -a00:00.0")
parser.add_argument("-T", "--timeout",
                    type=int,
                    help="Timeout in seconds",
                    default=600)
parser.add_argument("-c", "--test-cases",
                    nargs="+",
                    help="Defines test cases to run. Run all if not specified")
parser.add_argument("-v", "--test-vector",
                    nargs="+",
                    help="Specifies paths to the test vector files.",
                    default=[dpdk_path +
                    "/app/test-bbdev/test_vectors/bbdev_null.data"])
parser.add_argument("-n", "--num-ops",
                    type=int,
                    help="Number of operations to process on device.",
                    default=32)
parser.add_argument("-b", "--burst-size",
                    nargs="+",
                    type=int,
                    help="Operations enqueue/dequeue burst size.",
                    default=[32])
parser.add_argument("-s", "--snr",
                    type=int,
                    help="SNR in dB for BLER tests",
                    default=0)
parser.add_argument("-t", "--iter-max",
                    type=int,
                    help="Max iterations",
                    default=6)
parser.add_argument("-l", "--num-lcores",
                    type=int,
                    help="Number of lcores to run.",
                    default=16)
parser.add_argument("-i", "--init-device",
                    action='store_true',
                    help="Initialise PF device with default values.")

args = parser.parse_args()

if not os.path.exists(args.testapp_path):
    print("No such file: " + args.testapp_path)
    sys.exit(1)

params = [args.testapp_path]
if args.eal_params:
    params.extend(shlex.split(args.eal_params))

params.extend(["--"])

if args.snr:
    params.extend(["-s", str(args.snr)])

if args.iter_max:
    params.extend(["-t", str(args.iter_max)])

if args.num_ops:
    params.extend(["-n", str(args.num_ops)])

if args.num_lcores:
    params.extend(["-l", str(args.num_lcores)])

if args.test_cases:
    params.extend(["-c"])
    params.extend([",".join(args.test_cases)])

if args.init_device:
    params.extend(["-i"])


exit_status = 0
for vector in args.test_vector:
    for burst_size in args.burst_size:
        call_params = params[:]
        call_params.extend(["-v", vector])
        call_params.extend(["-b", str(burst_size)])
        params_string = " ".join(call_params)

        print("Executing: {}".format(params_string))
        try:
            output = subprocess.run(call_params, timeout=args.timeout, universal_newlines=True)
        except subprocess.TimeoutExpired as e:
            print("===========================================================")
            print("Starting Test Suite : BBdev TimeOut Tests")
            print("INFO: One of the tests timed out {}".format(e))
            print("INFO: Unexpected Error")
            print("+ ------------------------------------------------------- +")
            print("== test: timeout")
            print("Unexpected Error")
            print("TestCase [ 0] : timeout failed")
            print(" + ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ +")
            print(" + Tests Failed :       1")
            print(" + ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ +")
            exit_status = 1
        if output.returncode < 0:
            print("===========================================================")
            print("Starting Test Suite : BBdev Exception Tests")
            print("INFO: One of the tests returned {}".format(output.returncode))
            print("INFO: Unexpected Error")
            print("+ ------------------------------------------------------- +")
            print("== test: exception")
            print("Unexpected Error")
            print("TestCase [ 0] : exception failed")
            print(" + ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ +")
            print(" + Tests Failed :       1")
            print(" + ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ +")
            exit_status = 1
sys.exit(exit_status)



#! /usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2021 Intel Corporation

from sys import argv
from os.path import abspath

(h_file, c_file) = argv[1:]

contents = '#include "' + abspath(h_file) + '"'
with open(c_file, 'w') as cf:
    cf.write(contents)
